# Copyright (c) 2017  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import unittest
from os import path
from mock import patch
from module_build_service import models, conf
from tests import test_reuse_component_init_data, init_data, db
import mock
import koji
from module_build_service.scheduler.producer import MBSProducer
import six.moves.queue as queue
from datetime import datetime, timedelta

BASE_DIR = path.abspath(path.dirname(__file__))
CASSETTES_DIR = path.join(
    path.abspath(path.dirname(__file__)), '..', 'vcr-request-data')


@patch("module_build_service.builder.GenericBuilder.default_buildroot_groups",
       return_value={'build': [], 'srpm-build': []})
@patch("module_build_service.scheduler.consumer.get_global_consumer")
@patch("module_build_service.builder.KojiModuleBuilder.get_session")
@patch("module_build_service.builder.GenericBuilder.create_from_module")
class TestPoller(unittest.TestCase):

    def setUp(self):
        test_reuse_component_init_data()

    def tearDown(self):
        init_data()

    @patch('module_build_service.utils.start_build_component')
    def test_process_paused_module_builds(self, start_build_component, create_builder,
                                          koji_get_session, global_consumer,
                                          dbg):
        """
        Tests general use-case of process_paused_module_builds.
        """
        consumer = mock.MagicMock()
        consumer.incoming = queue.Queue()
        global_consumer.return_value = consumer

        koji_session = mock.MagicMock()
        koji_get_session.return_value = koji_session

        builder = mock.MagicMock()
        create_builder.return_value = builder

        # Change the batch to 2, so the module build is in state where
        # it is not building anything, but the state is "build".
        module_build = models.ModuleBuild.query.filter_by(id=2).one()
        module_build.batch = 2
        db.session.commit()

        # Poll :)
        hub = mock.MagicMock()
        poller = MBSProducer(hub)
        poller.poll()

        # Refresh our module_build object.
        module_build = models.ModuleBuild.query.filter_by(id=2).one()
        db.session.refresh(module_build)

        # Components should be in BUILDING state now.
        components = module_build.current_batch()
        for component in components:
            self.assertEqual(component.state, koji.BUILD_STATES["BUILDING"])
        self.assertEqual(len(start_build_component.mock_calls), 2)

    def test_trigger_new_repo_when_failed(self, create_builder,
                                          koji_get_session, global_consumer,
                                          dbg):
        """
        Tests that we call koji_sesion.newRepo when newRepo task failed.
        """
        consumer = mock.MagicMock()
        consumer.incoming = queue.Queue()
        global_consumer.return_value = consumer

        koji_session = mock.MagicMock()
        koji_session.getTag = lambda tag_name: {'name': tag_name}
        koji_session.getTaskInfo.return_value = {'state': koji.TASK_STATES['FAILED']}
        koji_session.newRepo.return_value = 123456
        koji_get_session.return_value = koji_session

        builder = mock.MagicMock()
        builder.buildroot_ready.return_value = False
        create_builder.return_value = builder

        # Change the batch to 2, so the module build is in state where
        # it is not building anything, but the state is "build".
        module_build = models.ModuleBuild.query.filter_by(id=2).one()
        module_build.batch = 2
        module_build.new_repo_task_id = 123456
        db.session.commit()

        hub = mock.MagicMock()
        poller = MBSProducer(hub)
        poller.poll()

        koji_session.newRepo.assert_called_once_with("module-testmodule-build")

    def test_trigger_new_repo_when_succeded(self, create_builder,
                                            koji_get_session, global_consumer,
                                            dbg):
        """
        Tests that we do not call koji_sesion.newRepo when newRepo task
        succeeded.
        """
        consumer = mock.MagicMock()
        consumer.incoming = queue.Queue()
        global_consumer.return_value = consumer

        koji_session = mock.MagicMock()
        koji_session.getTag = lambda tag_name: {'name': tag_name}
        koji_session.getTaskInfo.return_value = {'state': koji.TASK_STATES['CLOSED']}
        koji_session.newRepo.return_value = 123456
        koji_get_session.return_value = koji_session

        builder = mock.MagicMock()
        builder.buildroot_ready.return_value = False
        create_builder.return_value = builder

        # Change the batch to 2, so the module build is in state where
        # it is not building anything, but the state is "build".
        module_build = models.ModuleBuild.query.filter_by(id=2).one()
        module_build.batch = 2
        module_build.new_repo_task_id = 123456
        db.session.commit()

        hub = mock.MagicMock()
        poller = MBSProducer(hub)
        poller.poll()

        # Refresh our module_build object.
        module_build = models.ModuleBuild.query.filter_by(id=2).one()
        db.session.refresh(module_build)

        self.assertTrue(not koji_session.newRepo.called)
        self.assertEqual(module_build.new_repo_task_id, 0)

    def test_process_paused_module_builds_waiting_for_repo(
            self, create_builder, koji_get_session, global_consumer, dbg):
        """
        Tests that process_paused_module_builds does not start new batch
        when we are waiting for repo.
        """
        consumer = mock.MagicMock()
        consumer.incoming = queue.Queue()
        global_consumer.return_value = consumer

        koji_session = mock.MagicMock()
        koji_get_session.return_value = koji_session

        builder = mock.MagicMock()
        create_builder.return_value = builder

        # Change the batch to 2, so the module build is in state where
        # it is not building anything, but the state is "build".
        module_build = models.ModuleBuild.query.filter_by(id=2).one()
        module_build.batch = 2
        module_build.new_repo_task_id = 123456
        db.session.commit()

        # Poll :)
        hub = mock.MagicMock()
        poller = MBSProducer(hub)
        poller.poll()

        # Refresh our module_build object.
        module_build = models.ModuleBuild.query.filter_by(id=2).one()
        db.session.refresh(module_build)

        # Components should not be in building state
        components = module_build.current_batch()
        for component in components:
            self.assertEqual(component.state, None)

    def test_delete_old_koji_targets(
            self, create_builder, koji_get_session, global_consumer, dbg):
        """
        Tests that we delete koji target when time_completed is older than
        koji_target_delete_time value.
        """
        consumer = mock.MagicMock()
        consumer.incoming = queue.Queue()
        global_consumer.return_value = consumer

        for state_name, state in models.BUILD_STATES.items():
            koji_session = mock.MagicMock()
            koji_session.getBuildTargets.return_value = [
                {'dest_tag_name': 'module-tag', 'id': 852, 'name': 'module-tag'},
                {'dest_tag_name': 'f26', 'id': 853, 'name': 'f26'},
                {'dest_tag_name': 'module-tag2', 'id': 853, 'name': 'f26'}]
            koji_get_session.return_value = koji_session

            builder = mock.MagicMock()
            create_builder.return_value = builder

            # Change the batch to 2, so the module build is in state where
            # it is not building anything, but the state is "build".
            module_build = models.ModuleBuild.query.filter_by(id=2).one()
            module_build.state = state
            module_build.koji_tag = "module-tag"
            module_build.time_completed = datetime.utcnow()
            module_build.new_repo_task_id = 123456
            db.session.commit()

            # Poll :)
            hub = mock.MagicMock()
            poller = MBSProducer(hub)
            poller.delete_old_koji_targets(conf, db.session)

            module_build = models.ModuleBuild.query.filter_by(id=2).one()
            db.session.refresh(module_build)
            module_build.time_completed = datetime.utcnow() - timedelta(hours=23)
            db.session.commit()
            poller.delete_old_koji_targets(conf, db.session)

            # deleteBuildTarget should not be called, because time_completed is
            # set to "now".
            self.assertTrue(not koji_session.deleteBuildTarget.called)

            # Try removing non-modular target - should not happen
            db.session.refresh(module_build)
            module_build.koji_tag = "module-tag2"
            module_build.time_completed = datetime.utcnow() - timedelta(hours=25)
            db.session.commit()
            poller.delete_old_koji_targets(conf, db.session)
            self.assertTrue(not koji_session.deleteBuildTarget.called)

            # Refresh our module_build object and set time_completed 25 hours ago
            db.session.refresh(module_build)
            module_build.time_completed = datetime.utcnow() - timedelta(hours=25)
            module_build.koji_tag = "module-tag"
            db.session.commit()

            poller.delete_old_koji_targets(conf, db.session)

            if state_name in ["done", "ready", "failed"]:
                koji_session.deleteBuildTarget.assert_called_once_with(852)

    def test_process_waiting_module_build(self, create_builder, koji_get_session,
                                          global_consumer, dbg):
        """ Test that processing old waiting module builds works. """

        consumer = mock.MagicMock()
        consumer.incoming = queue.Queue()
        global_consumer.return_value = consumer

        hub = mock.MagicMock()
        poller = MBSProducer(hub)

        # Change the batch to 2, so the module build is in state where
        # it is not building anything, but the state is "build".
        module_build = models.ModuleBuild.query.filter_by(id=2).one()
        module_build.state = 1
        original = datetime.utcnow() - timedelta(minutes=11)
        module_build.time_modified = original
        db.session.commit()
        db.session.refresh(module_build)

        # Ensure the queue is empty before we start.
        self.assertEquals(consumer.incoming.qsize(), 0)

        # Poll :)
        poller.process_waiting_module_builds(db.session)

        self.assertEquals(consumer.incoming.qsize(), 1)
        module_build = models.ModuleBuild.query.filter_by(id=2).one()
        # ensure the time_modified was changed.
        self.assertGreater(module_build.time_modified, original)

    def test_process_waiting_module_build_not_old_enough(self, create_builder, koji_get_session,
                                                         global_consumer, dbg):
        """ Test that we do not process young waiting builds. """

        consumer = mock.MagicMock()
        consumer.incoming = queue.Queue()
        global_consumer.return_value = consumer

        hub = mock.MagicMock()
        poller = MBSProducer(hub)

        # Change the batch to 2, so the module build is in state where
        # it is not building anything, but the state is "build".
        module_build = models.ModuleBuild.query.filter_by(id=2).one()
        module_build.state = 1
        original = datetime.utcnow() - timedelta(minutes=9)
        module_build.time_modified = original
        db.session.commit()
        db.session.refresh(module_build)

        # Ensure the queue is empty before we start.
        self.assertEquals(consumer.incoming.qsize(), 0)

        # Poll :)
        poller.process_waiting_module_builds(db.session)

        # Ensure we did *not* process the 9 minute-old build.
        self.assertEquals(consumer.incoming.qsize(), 0)

    def test_process_waiting_module_build_none_found(self, create_builder, koji_get_session,
                                                     global_consumer, dbg):
        """ Test nothing happens when no module builds are waiting. """

        consumer = mock.MagicMock()
        consumer.incoming = queue.Queue()
        global_consumer.return_value = consumer

        hub = mock.MagicMock()
        poller = MBSProducer(hub)

        # Ensure the queue is empty before we start.
        self.assertEquals(consumer.incoming.qsize(), 0)

        # Poll :)
        poller.process_waiting_module_builds(db.session)

        # Ensure we did *not* process any of the non-waiting builds.
        self.assertEquals(consumer.incoming.qsize(), 0)



# Copyright (c) 2017  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Written by Jan Kaluza <jkaluza@redhat.com>


import os
import shutil
import tempfile
import unittest

from module_build_service import log
from module_build_service.logger import ModuleBuildLogs
from module_build_service.scheduler.consumer import MBSConsumer


class TestLogger(unittest.TestCase):

    def setUp(self):
        self.base = tempfile.mkdtemp(prefix='mbs-', suffix='-%s' % self.id())
        print("Storing build logs in %r" % self.base)
        self.build_log = ModuleBuildLogs(self.base)

    def tearDown(self):
        MBSConsumer.current_module_build_id = None
        shutil.rmtree(self.base)

    def test_module_build_logs(self):
        """
        Tests that ModuleBuildLogs is logging properly to build log file.
        """
        # Initialize logging, get the build log path and remove it to
        # ensure we are not using some garbage from previous failed test.
        self.build_log.start(1)
        path = self.build_log.path(1)
        self.assertEqual(path[len(self.base):], "/build-1.log")
        if os.path.exists(path):
            os.unlink(path)

        # Try logging without the MBSConsumer.current_module_build_id set.
        # No log file should be created.
        log.debug("ignore this test msg")
        log.info("ignore this test msg")
        log.warn("ignore this test msg")
        log.error("ignore this test msg")
        self.build_log.stop(1)
        self.assertTrue(not os.path.exists(path))

        # Try logging with current_module_build_id set to 1 and then to 2.
        # Only messages with current_module_build_id set to 1 should appear in
        # the log.
        self.build_log.start(1)
        MBSConsumer.current_module_build_id = 1
        log.debug("ignore this test msg1")
        log.info("ignore this test msg1")
        log.warn("ignore this test msg1")
        log.error("ignore this test msg1")

        MBSConsumer.current_module_build_id = 2
        log.debug("ignore this test msg2")
        log.info("ignore this test msg2")
        log.warn("ignore this test msg2")
        log.error("ignore this test msg2")

        self.build_log.stop(1)
        self.assertTrue(os.path.exists(path))
        with open(path, "r") as f:
            data = f.read()
            for level in ["DEBUG", "INFO", "WARNING", "ERROR"]:
                self.assertTrue(data.find("%s - ignore this test msg1" % level) != -1)

        # Try to log more messages when build_log for module 1 is stopped.
        # New messages should not appear in a log.
        MBSConsumer.current_module_build_id = 1
        log.debug("ignore this test msg3")
        log.info("ignore this test msg3")
        log.warn("ignore this test msg3")
        log.error("ignore this test msg3")
        self.build_log.stop(1)
        with open(path, "r") as f:
            data = f.read()
            self.assertTrue(data.find("ignore this test msg3") == -1)



# Copyright (c) 2017  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import unittest
from os import path, mkdir
from shutil import copyfile
import vcr
import modulemd
from mock import patch
import module_build_service.utils
import module_build_service.scm
from module_build_service import models, conf
from module_build_service.errors import ProgrammingError, ValidationError
from tests import (test_reuse_component_init_data, init_data, db,
                   test_reuse_shared_userspace_init_data)
import mock
import koji
import module_build_service.scheduler.handlers.components
from module_build_service.builder import GenericBuilder, KojiModuleBuilder
from tests import app

BASE_DIR = path.abspath(path.dirname(__file__))
CASSETTES_DIR = path.join(
    path.abspath(path.dirname(__file__)), '..', 'vcr-request-data')


class FakeSCM(object):
    def __init__(self, mocked_scm, name, mmd_filename, commit=None):
        self.mocked_scm = mocked_scm
        self.name = name
        self.commit = commit
        self.mmd_filename = mmd_filename
        self.sourcedir = None

        self.mocked_scm.return_value.checkout = self.checkout
        self.mocked_scm.return_value.name = self.name
        self.mocked_scm.return_value.branch = 'master'
        self.mocked_scm.return_value.get_latest = self.get_latest
        self.mocked_scm.return_value.commit = self.commit
        self.mocked_scm.return_value.repository_root = "git://pkgs.stg.fedoraproject.org/modules/"
        self.mocked_scm.return_value.sourcedir = self.sourcedir
        self.mocked_scm.return_value.get_module_yaml = self.get_module_yaml

    def checkout(self, temp_dir):
        self.sourcedir = path.join(temp_dir, self.name)
        mkdir(self.sourcedir)
        base_dir = path.abspath(path.dirname(__file__))
        copyfile(path.join(base_dir, '..', 'staged_data', self.mmd_filename),
                 self.get_module_yaml())

        return self.sourcedir

    def get_latest(self, branch='master'):
        return self.commit if self.commit else branch

    def get_module_yaml(self):
        return path.join(self.sourcedir, self.name + ".yaml")


class TestUtils(unittest.TestCase):

    def setUp(self):
        self.filtered_rpms = [
            u'glibc-utils-0:2.25-4.module_5ccf9229',
            u'glibc-benchtests-0:2.25-4.module_5ccf9229',
            u'systemd-journal-remote-0:233-3.module_5ccf9229',
            u'openldap-servers-0:2.4.44-8.module_5ccf9229',
            u'kernel-debug-devel-0:4.11.0-0.rc7.git0.1.module_5ccf9229',
            u'python-perf-0:4.11.0-0.rc7.git0.1.module_5ccf9229',
            u'kernel-tools-0:4.11.0-0.rc7.git0.1.module_5ccf9229',
            u'perf-0:4.11.0-0.rc7.git0.1.module_5ccf9229',
            u'kernel-tools-libs-devel-0:4.11.0-0.rc7.git0.1.module_5ccf9229',
            u'kernel-devel-0:4.11.0-0.rc7.git0.1.module_5ccf9229',
            u'kernel-PAE-devel-0:4.11.0-0.rc7.git0.1.module_5ccf9229',
            u'kernel-PAEdebug-devel-0:4.11.0-0.rc7.git0.1.module_5ccf9229',
            u'kernel-lpae-devel-0:4.11.0-0.rc7.git0.1.module_5ccf9229',
            u'libcroco-devel-0:0.6.11-3.module_5ccf9229',
            u'dbus-x11-1:1.11.10-2.module_5ccf9229',
            u'qgpgme-0:1.9.0-1.module_5ccf9229',
            u'qgpgme-devel-0:1.9.0-1.module_5ccf9229',
            u'python2-gpg-0:1.9.0-1.module_5ccf9229',
            u'libssh2-devel-0:1.8.0-2.module_5ccf9229',
            u'python3-test-0:3.6.0-21.module_5ccf9229',
            u'python3-debug-0:3.6.0-21.module_5ccf9229',
            u'python3-tkinter-0:3.6.0-21.module_5ccf9229',
            u'python3-tools-0:3.6.0-21.module_5ccf9229',
            u'grub2-starfield-theme-1:2.02-0.38.module_5ccf9229',
            u'libbabeltrace-devel-0:1.5.2-2.module_5ccf9229',
            u'libidn-javadoc-0:1.33-2.module_5ccf9229',
            u'libidn-java-0:1.33-2.module_5ccf9229',
            u'cyrus-sasl-sql-0:2.1.26-30.module_5ccf9229',
            u'python2-solv-0:0.6.26-1.module_5ccf9229',
            u'perl-solv-0:0.6.26-1.module_5ccf9229',
            u'python2-packaging-0:16.8-4.module_5ccf9229',
            u'freetype-demos-0:2.7.1-2.module_5ccf9229',
            u'util-linux-user-0:2.29.1-2.module_5ccf9229',
            u'syslinux-perl-0:6.04-0.2.module_5ccf9229',
            u'gnutls-utils-0:3.5.10-1.module_5ccf9229',
            u'gnutls-guile-0:3.5.10-1.module_5ccf9229',
            u'gnutls-devel-0:3.5.10-1.module_5ccf9229',
            u'gnutls-dane-0:3.5.10-1.module_5ccf9229',
            u'python-pwquality-0:1.3.0-8.module_5ccf9229',
            u'openssl-perl-1:1.1.0e-1.module_5ccf9229',
            u'glib2-fam-0:2.52.0-1.module_5ccf9229',
            u'glib2-static-0:2.52.0-1.module_5ccf9229',
            u'glib2-devel-0:2.52.0-1.module_5ccf9229',
            u'libselinux-ruby-0:2.6-2.module_5ccf9229',
            u'libselinux-python-0:2.6-2.module_5ccf9229',
            u'cmirror-0:2.02.168-4.module_5ccf9229',
            u'lvm2-cluster-standalone-0:2.02.168-4.module_5ccf9229',
            u'lvm2-dbusd-0:2.02.168-4.module_5ccf9229',
            u'lvm2-python-libs-0:2.02.168-4.module_5ccf9229',
            u'lvm2-cluster-0:2.02.168-4.module_5ccf9229',
            u'cmirror-standalone-0:2.02.168-4.module_5ccf9229',
            u'lvm2-lockd-0:2.02.168-4.module_5ccf9229',
            u'python2-pip-0:9.0.1-7.module_5ccf9229',
            u'python-magic-0:5.30-5.module_5ccf9229',
            u'cracklib-python-0:2.9.6-5.module_5ccf9229',
            u'hfsutils-x11-0:3.2.6-31.module_5ccf9229',
            u'kernel-rpm-macros-0:63-1.module_5ccf9229',
            u'cryptsetup-python-0:1.7.3-3.module_5ccf9229',
            u'python2-rpm-0:4.13.0.1-3.module_5ccf9229',
            u'rpm-cron-0:4.13.0.1-3.module_5ccf9229',
            u'texinfo-0:6.3-2.module_5ccf9229',
            u'texinfo-tex-0:6.3-2.module_5ccf9229',
            u'qt5-0:5.8.0-2.module_5ccf9229',
            u'qt5-devel-0:5.8.0-2.module_5ccf9229',
            u'qt5-rpm-macros-0:5.8.0-2.module_5ccf9229',
            u'gnupg2-smime-0:2.1.18-2.module_5ccf9229',
            u'emacs-nox-1:25.2-0.1.rc2.module_5ccf9229',
            u'emacs-terminal-1:25.2-0.1.rc2.module_5ccf9229',
            u'emacs-common-1:25.2-0.1.rc2.module_5ccf9229',
            u'emacs-1:25.2-0.1.rc2.module_5ccf9229',
            u'libverto-libevent-devel-0:0.2.6-7.module_5ccf9229',
            u'libverto-tevent-devel-0:0.2.6-7.module_5ccf9229',
            u'libverto-libevent-0:0.2.6-7.module_5ccf9229',
            u'libverto-tevent-0:0.2.6-7.module_5ccf9229',
            u'libtool-0:2.4.6-17.module_5ccf9229',
            u'libpeas-loader-python3-0:1.20.0-5.module_5ccf9229',
            u'libpeas-devel-0:1.20.0-5.module_5ccf9229',
            u'libpeas-gtk-0:1.20.0-5.module_5ccf9229',
            u'libpeas-loader-python-0:1.20.0-5.module_5ccf9229',
            u'python2-setuptools-0:34.3.0-1.module_5ccf9229',
            u'libsemanage-python-0:2.6-2.module_5ccf9229',
            u'python-libxml2-0:2.9.4-2.module_5ccf9229',
            u'krb5-server-ldap-0:1.15-9.module_5ccf9229',
            u'krb5-server-0:1.15-9.module_5ccf9229',
            u'python2-six-0:1.10.0-8.module_5ccf9229',
            u'python2-appdirs-0:1.4.0-10.module_5ccf9229',
            u'pyparsing-0:2.1.10-3.module_5ccf9229',
            u'python2-pyparsing-0:2.1.10-3.module_5ccf9229',
            u'ntsysv-0:1.9-1.module_5ccf9229',
            u'audit-0:2.7.3-1.module_5ccf9229',
            u'audit-libs-python-0:2.7.3-1.module_5ccf9229',
            u'audispd-plugins-0:2.7.3-1.module_5ccf9229',
            u'audispd-plugins-zos-0:2.7.3-1.module_5ccf9229',
            u'audit-libs-python3-0:2.7.3-1.module_5ccf9229',
            u'modeline2fb-0:2.1-40.module_5ccf9229',
            u'msghack-0:0.19.8.1-8.module_5ccf9229',
            u'emacs-gettext-0:0.19.8.1-8.module_5ccf9229',
            u'python2-hawkey-0:0.8.2-1.module_987f08f4',
            u'libdnf-devel-0:0.8.2-1.module_987f08f4',
            u'sqlite-tcl-0:3.17.0-2.module_5ccf9229',
            u'sqlite-analyzer-0:3.17.0-2.module_5ccf9229',
            u'sssd-0:1.15.2-4.module_47fecbcd',
            u'sssd-nfs-idmap-0:1.15.2-4.module_47fecbcd',
            u'sssd-winbind-idmap-0:1.15.2-4.module_47fecbcd',
            u'sssd-krb5-0:1.15.2-4.module_47fecbcd',
            u'libsss_sudo-0:1.15.2-4.module_47fecbcd',
            u'sssd-libwbclient-devel-0:1.15.2-4.module_47fecbcd',
            u'sssd-common-pac-0:1.15.2-4.module_47fecbcd',
            u'sssd-ipa-0:1.15.2-4.module_47fecbcd',
            u'python3-sss-murmur-0:1.15.2-4.module_47fecbcd',
            u'libsss_simpleifp-devel-0:1.15.2-4.module_47fecbcd',
            u'sssd-common-0:1.15.2-4.module_47fecbcd',
            u'sssd-dbus-0:1.15.2-4.module_47fecbcd',
            u'libipa_hbac-0:1.15.2-4.module_47fecbcd',
            u'sssd-proxy-0:1.15.2-4.module_47fecbcd',
            u'python2-sss-murmur-0:1.15.2-4.module_47fecbcd',
            u'python3-sss-0:1.15.2-4.module_47fecbcd',
            u'libsss_autofs-0:1.15.2-4.module_47fecbcd',
            u'python2-libipa_hbac-0:1.15.2-4.module_47fecbcd',
            u'libsss_simpleifp-0:1.15.2-4.module_47fecbcd',
            u'libsss_idmap-devel-0:1.15.2-4.module_47fecbcd',
            u'sssd-libwbclient-0:1.15.2-4.module_47fecbcd',
            u'python2-libsss_nss_idmap-0:1.15.2-4.module_47fecbcd',
            u'python3-libipa_hbac-0:1.15.2-4.module_47fecbcd',
            u'python2-sssdconfig-0:1.15.2-4.module_47fecbcd',
            u'libipa_hbac-devel-0:1.15.2-4.module_47fecbcd',
            u'libsss_nss_idmap-devel-0:1.15.2-4.module_47fecbcd',
            u'python2-sss-0:1.15.2-4.module_47fecbcd',
            u'sssd-ldap-0:1.15.2-4.module_47fecbcd',
            u'sssd-ad-0:1.15.2-4.module_47fecbcd',
            u'sssd-tools-0:1.15.2-4.module_47fecbcd',
            u'sssd-krb5-common-0:1.15.2-4.module_47fecbcd',
            u'librepo-devel-0:1.7.20-3.module_5ccf9229',
            u'python2-librepo-0:1.7.20-3.module_5ccf9229',
            u'iproute-tc-0:4.11.0-1.module_d6de39f1',
            u'dracut-live-0:044-182.module_bd7491c8',
            u'dracut-fips-aesni-0:044-182.module_bd7491c8',
            u'dracut-fips-0:044-182.module_bd7491c8',
            u'dracut-network-0:044-182.module_bd7491c8',
            u'gobject-introspection-devel-0:1.52.0-1.module_5ccf9229',
            u'iptables-compat-0:1.6.1-2.module_5ccf9229',
            u'libcap-ng-python-0:0.7.8-3.module_5ccf9229',
            u'python3-sssdconfig-0:1.15.2-4.module_47fecbcd'
        ]

    def tearDown(self):
        init_data()

    @vcr.use_cassette(
        path.join(CASSETTES_DIR, 'tests.test_utils.TestUtils.test_format_mmd'))
    @patch('module_build_service.scm.SCM')
    def test_format_mmd(self, mocked_scm):
        mocked_scm.return_value.commit = \
            '620ec77321b2ea7b0d67d82992dda3e1d67055b4'
        # For all the RPMs in testmodule, get_latest is called
        hashes_returned = {
            'f24': '4ceea43add2366d8b8c5a622a2fb563b625b9abf',
            'f23': 'fbed359411a1baa08d4a88e0d12d426fbf8f602c',
            'f25': '76f9d8c8e87eed0aab91034b01d3d5ff6bd5b4cb'}
        original_refs = ["f23", "f24", "f25"]

        def mocked_get_latest(branch="master"):
            return hashes_returned[branch]

        mocked_scm.return_value.get_latest = mocked_get_latest
        mmd = modulemd.ModuleMetadata()
        with open(path.join(BASE_DIR, '..', 'staged_data', 'testmodule.yaml')) \
                as mmd_file:
            mmd.loads(mmd_file)
        scmurl = \
            ('git://pkgs.stg.fedoraproject.org/modules/testmodule.git'
             '?#620ec77321b2ea7b0d67d82992dda3e1d67055b4')
        module_build_service.utils.format_mmd(mmd, scmurl)

        # Make sure that original refs are not changed.
        mmd_pkg_refs = [pkg.ref for pkg in mmd.components.rpms.values()]
        self.assertEqual(set(mmd_pkg_refs), set(original_refs))

        self.assertEqual(mmd.buildrequires, {'base-runtime': 'master'})
        xmd = {
            'mbs': {
                'commit': '620ec77321b2ea7b0d67d82992dda3e1d67055b4',
                'buildrequires': {
                    'base-runtime': {
                        'ref': '147dca4ca65aa9a1ac51f71b7e687f9178ffa5df',
                        'stream': 'master',
                        'version': '20170616125652',
                        'filtered_rpms': self.filtered_rpms}},
                'requires': {
                    'base-runtime': {
                        'version': '20170616125652',
                        'ref': '147dca4ca65aa9a1ac51f71b7e687f9178ffa5df',
                        'stream': 'master',
                        'filtered_rpms': self.filtered_rpms}},
                'rpms': {'perl-List-Compare': {'ref': '76f9d8c8e87eed0aab91034b01d3d5ff6bd5b4cb'},
                         'perl-Tangerine': {'ref': '4ceea43add2366d8b8c5a622a2fb563b625b9abf'},
                         'tangerine': {'ref': 'fbed359411a1baa08d4a88e0d12d426fbf8f602c'}},
                'scmurl': 'git://pkgs.stg.fedoraproject.org/modules/testmodule'
                          '.git?#620ec77321b2ea7b0d67d82992dda3e1d67055b4',
            }
        }

        self.assertEqual(mmd.xmd, xmd)

    @vcr.use_cassette(
        path.join(CASSETTES_DIR, 'tests.test_utils.TestUtils.test_format_mmd'))
    @patch('module_build_service.scm.SCM')
    def test_format_mmd_empty_scmurl(self, mocked_scm):
        # For all the RPMs in testmodule, get_latest is called
        hashes_returned = {
            'f24': '4ceea43add2366d8b8c5a622a2fb563b625b9abf',
            'f23': 'fbed359411a1baa08d4a88e0d12d426fbf8f602c',
            'f25': '76f9d8c8e87eed0aab91034b01d3d5ff6bd5b4cb'}

        def mocked_get_latest(branch="master"):
            return hashes_returned[branch]
        mocked_scm.return_value.get_latest = mocked_get_latest

        mmd = modulemd.ModuleMetadata()
        with open(path.join(BASE_DIR, '..', 'staged_data', 'testmodule.yaml')) \
                as mmd_file:
            mmd.loads(mmd_file)

        module_build_service.utils.format_mmd(mmd, scmurl=None)
        xmd = {
            'mbs': {
                'commit': None,
                'buildrequires': {
                    'base-runtime': {
                        'ref': '147dca4ca65aa9a1ac51f71b7e687f9178ffa5df',
                        'stream': 'master',
                        'version': '20170616125652',
                        'filtered_rpms': self.filtered_rpms}},
                'requires': {
                    'base-runtime': {
                        'version': '20170616125652',
                        'ref': '147dca4ca65aa9a1ac51f71b7e687f9178ffa5df',
                        'stream': 'master',
                        'filtered_rpms': self.filtered_rpms}},
                'rpms': {'perl-List-Compare': {'ref': '76f9d8c8e87eed0aab91034b01d3d5ff6bd5b4cb'},
                         'perl-Tangerine': {'ref': '4ceea43add2366d8b8c5a622a2fb563b625b9abf'},
                         'tangerine': {'ref': 'fbed359411a1baa08d4a88e0d12d426fbf8f602c'}},
                'scmurl': None,
            }
        }
        self.assertEqual(mmd.xmd, xmd)

    def test_get_reusable_component_same(self):
        test_reuse_component_init_data()
        new_module = models.ModuleBuild.query.filter_by(id=2).one()
        rv = module_build_service.utils.get_reusable_component(
            db.session, new_module, 'tangerine')
        self.assertEqual(rv.package, 'tangerine')

    def test_get_reusable_component_empty_scmurl(self):
        test_reuse_component_init_data()

        new_module = models.ModuleBuild.query.filter_by(id=2).one()
        mmd = new_module.mmd()
        mmd.xmd['mbs']['buildrequires'] = {'base-runtime': {}}
        new_module.modulemd = mmd.dumps()
        db.session.commit()

        rv = module_build_service.utils.get_reusable_component(
            db.session, new_module, 'tangerine')
        self.assertEqual(rv, None)

    def test_get_reusable_component_different_perl_tangerine(self):
        test_reuse_component_init_data()
        second_module_build = models.ModuleBuild.query.filter_by(id=2).one()
        mmd = second_module_build.mmd()
        mmd.components.rpms['perl-Tangerine'].ref = \
            '00ea1da4192a2030f9ae023de3b3143ed647bbab'
        second_module_build.modulemd = mmd.dumps()
        second_module_perl_tangerine = models.ComponentBuild.query.filter_by(
            package='perl-Tangerine', module_id=2).one()
        second_module_perl_tangerine.ref = \
            '00ea1da4192a2030f9ae023de3b3143ed647bbab'
        db.session.commit()
        # Shares the same build order as the changed perl-Tangerine, but none
        # of the build orders before it are different (in this case there are
        # none)
        plc_rv = module_build_service.utils.get_reusable_component(
            db.session, second_module_build, 'perl-List-Compare')
        self.assertEqual(plc_rv.package, 'perl-List-Compare')

        # perl-Tangerine has a different commit hash
        pt_rv = module_build_service.utils.get_reusable_component(
            db.session, second_module_build, 'perl-Tangerine')
        self.assertEqual(pt_rv, None)

        # tangerine is the same but its in a build order that is after the
        # different perl-Tangerine, so it can't be reused
        tangerine_rv = module_build_service.utils.get_reusable_component(
            db.session, second_module_build, 'tangerine')
        self.assertEqual(tangerine_rv, None)

    def test_get_reusable_component_different_rpm_macros(self):
        test_reuse_component_init_data()
        second_module_build = models.ModuleBuild.query.filter_by(id=2).one()
        mmd = second_module_build.mmd()
        mmd.buildopts.rpms.macros = "%my_macro 1"
        second_module_build.modulemd = mmd.dumps()
        db.session.commit()

        plc_rv = module_build_service.utils.get_reusable_component(
            db.session, second_module_build, 'perl-List-Compare')
        self.assertEqual(plc_rv, None)

        # perl-Tangerine has a different commit hash
        pt_rv = module_build_service.utils.get_reusable_component(
            db.session, second_module_build, 'perl-Tangerine')
        self.assertEqual(pt_rv, None)

    def test_get_reusable_component_different_buildrequires_hash(self):
        test_reuse_component_init_data()
        second_module_build = models.ModuleBuild.query.filter_by(id=2).one()
        mmd = second_module_build.mmd()
        mmd.xmd['mbs']['buildrequires']['base-runtime']['ref'] = \
            'da39a3ee5e6b4b0d3255bfef95601890afd80709'
        second_module_build.modulemd = mmd.dumps()
        db.session.commit()

        plc_rv = module_build_service.utils.get_reusable_component(
            db.session, second_module_build, 'perl-List-Compare')
        self.assertEqual(plc_rv, None)

        # perl-Tangerine has a different commit hash
        pt_rv = module_build_service.utils.get_reusable_component(
            db.session, second_module_build, 'perl-Tangerine')
        self.assertEqual(pt_rv, None)

        # tangerine is the same but its in a build order that is after the
        # different perl-Tangerine, so it can't be reused
        tangerine_rv = module_build_service.utils.get_reusable_component(
            db.session, second_module_build, 'tangerine')
        self.assertEqual(tangerine_rv, None)

    def test_get_reusable_component_different_buildrequires(self):
        test_reuse_component_init_data()
        second_module_build = models.ModuleBuild.query.filter_by(id=2).one()
        mmd = second_module_build.mmd()
        mmd.buildrequires = {'some_module': 'master'}
        mmd.xmd['mbs']['buildrequires'] = {
            'some_module': {
                'ref': 'da39a3ee5e6b4b0d3255bfef95601890afd80709',
                'stream': 'master',
                'version': '20170123140147'
            }
        }
        second_module_build.modulemd = mmd.dumps()
        db.session.commit()

        plc_rv = module_build_service.utils.get_reusable_component(
            db.session, second_module_build, 'perl-List-Compare')
        self.assertEqual(plc_rv, None)

        # perl-Tangerine has a different commit hash
        pt_rv = module_build_service.utils.get_reusable_component(
            db.session, second_module_build, 'perl-Tangerine')
        self.assertEqual(pt_rv, None)

        # tangerine is the same but its in a build order that is after the
        # different perl-Tangerine, so it can't be reused
        tangerine_rv = module_build_service.utils.get_reusable_component(
            db.session, second_module_build, 'tangerine')
        self.assertEqual(tangerine_rv, None)

    def test_get_reusable_component_shared_userspace_ordering(self):
        """
        For modules with lot of components per batch, there is big chance that
        the database will return them in different order than what we have for
        current `new_module`. In this case, reuse code should still be able to
        reuse the components.
        """
        test_reuse_shared_userspace_init_data()
        new_module = models.ModuleBuild.query.filter_by(id=2).one()
        rv = module_build_service.utils.get_reusable_component(
            db.session, new_module, 'llvm')
        self.assertEqual(rv.package, 'llvm')

    def test_validate_koji_tag_wrong_tag_arg_during_programming(self):
        """ Test that we fail on a wrong param name (non-existing one) due to
        programming error. """

        @module_build_service.utils.validate_koji_tag('wrong_tag_arg')
        def validate_koji_tag_programming_error(good_tag_arg, other_arg):
            pass

        with self.assertRaises(ProgrammingError):
            validate_koji_tag_programming_error('dummy', 'other_val')

    def test_validate_koji_tag_bad_tag_value(self):
        """ Test that we fail on a bad tag value. """

        @module_build_service.utils.validate_koji_tag('tag_arg')
        def validate_koji_tag_bad_tag_value(tag_arg):
            pass

        with self.assertRaises(ValidationError):
            validate_koji_tag_bad_tag_value('forbiddentagprefix-foo')

    def test_validate_koji_tag_bad_tag_value_in_list(self):
        """ Test that we fail on a list containing bad tag value. """

        @module_build_service.utils.validate_koji_tag('tag_arg')
        def validate_koji_tag_bad_tag_value_in_list(tag_arg):
            pass

        with self.assertRaises(ValidationError):
            validate_koji_tag_bad_tag_value_in_list([
                'module-foo', 'forbiddentagprefix-bar'])

    def test_validate_koji_tag_good_tag_value(self):
        """ Test that we pass on a good tag value. """

        @module_build_service.utils.validate_koji_tag('tag_arg')
        def validate_koji_tag_good_tag_value(tag_arg):
            return True

        self.assertEquals(
            validate_koji_tag_good_tag_value('module-foo'), True)

    def test_validate_koji_tag_good_tag_values_in_list(self):
        """ Test that we pass on a list of good tag values. """

        @module_build_service.utils.validate_koji_tag('tag_arg')
        def validate_koji_tag_good_tag_values_in_list(tag_arg):
            return True

        self.assertEquals(
            validate_koji_tag_good_tag_values_in_list(['module-foo',
                                                       'module-bar']), True)

    def test_validate_koji_tag_good_tag_value_in_dict(self):
        """ Test that we pass on a dict arg with default key
        and a good value. """

        @module_build_service.utils.validate_koji_tag('tag_arg')
        def validate_koji_tag_good_tag_value_in_dict(tag_arg):
            return True

        self.assertEquals(
            validate_koji_tag_good_tag_value_in_dict({'name': 'module-foo'}), True)

    def test_validate_koji_tag_good_tag_value_in_dict_nondefault_key(self):
        """ Test that we pass on a dict arg with non-default key
        and a good value. """

        @module_build_service.utils.validate_koji_tag('tag_arg',
                                                      dict_key='nondefault')
        def validate_koji_tag_good_tag_value_in_dict_nondefault_key(tag_arg):
            return True

        self.assertEquals(
            validate_koji_tag_good_tag_value_in_dict_nondefault_key(
                {'nondefault': 'module-foo'}), True)

    def test_validate_koji_tag_double_trouble_good(self):
        """ Test that we pass on a list of tags that are good. """

        expected = 'foo'

        @module_build_service.utils.validate_koji_tag(['tag_arg1', 'tag_arg2'])
        def validate_koji_tag_double_trouble(tag_arg1, tag_arg2):
            return expected

        actual = validate_koji_tag_double_trouble('module-1', 'module-2')
        self.assertEquals(actual, expected)

    def test_validate_koji_tag_double_trouble_bad(self):
        """ Test that we fail on a list of tags that are bad. """

        @module_build_service.utils.validate_koji_tag(['tag_arg1', 'tag_arg2'])
        def validate_koji_tag_double_trouble(tag_arg1, tag_arg2):
            pass

        with self.assertRaises(ValidationError):
            validate_koji_tag_double_trouble('module-1', 'BADNEWS-2')

    def test_validate_koji_tag_is_None(self):
        """ Test that we fail on a tag which is None. """

        @module_build_service.utils.validate_koji_tag('tag_arg')
        def validate_koji_tag_is_None(tag_arg):
            pass

        with self.assertRaises(ValidationError) as cm:
            validate_koji_tag_is_None(None)

        self.assertTrue(str(cm.exception).endswith(' No value provided.'))

    @vcr.use_cassette(
        path.join(CASSETTES_DIR, 'tests.test_utils.TestUtils.test_format_mmd'))
    @patch('module_build_service.scm.SCM')
    def test_resubmit(self, mocked_scm):
        """
        Tests that the module resubmit reintializes the module state and
        component states properly.
        """
        FakeSCM(mocked_scm, 'testmodule', 'testmodule.yaml',
                '620ec77321b2ea7b0d67d82992dda3e1d67055b4')
        with app.app_context():
            test_reuse_component_init_data()
            # Mark the module build as failed, so we can resubmit it.
            module_build = models.ModuleBuild.query.filter_by(id=2).one()
            module_build.batch = 2
            module_build.state = models.BUILD_STATES['failed']
            module_build.state_reason = "Cancelled"
            module_build.version = 1

            # Mark the components as COMPLETE/FAILED/CANCELED
            components = module_build.component_builds
            complete_component = components[0]
            complete_component.state = koji.BUILD_STATES['COMPLETE']
            failed_component = components[1]
            failed_component.state = koji.BUILD_STATES['FAILED']
            canceled_component = components[2]
            canceled_component.state = koji.BUILD_STATES['CANCELED']
            db.session.commit()

            module_build_service.utils.submit_module_build_from_scm(
                "Tom Brady", 'git://pkgs.stg.fedoraproject.org/modules/testmodule.git?#8fea453',
                'master')

            self.assertEqual(module_build.state, models.BUILD_STATES['wait'])
            self.assertEqual(module_build.batch, 0)
            self.assertEqual(module_build.state_reason, "Resubmitted by Tom Brady")
            self.assertEqual(complete_component.state, koji.BUILD_STATES['COMPLETE'])
            # The failed/cancelled components are now stateless
            self.assertIsNone(failed_component.state)
            self.assertIsNone(canceled_component.state)

    @vcr.use_cassette(
        path.join(CASSETTES_DIR, ('tests.test_utils.TestUtils.'
                                  'test_record_component_builds_duplicate_components')))
    @patch('module_build_service.scm.SCM')
    def test_record_component_builds_duplicate_components(self, mocked_scm):
        with app.app_context():
            test_reuse_component_init_data()
            mocked_scm.return_value.commit = \
                '620ec77321b2ea7b0d67d82992dda3e1d67055b4'
            # For all the RPMs in testmodule, get_latest is called
            hashes_returned = {
                'f25': '4ceea43add2366d8b8c5a622a2fb563b625b9abf',
                'f24': 'fbed359411a1baa08d4a88e0d12d426fbf8f602c'}

            def mocked_get_latest(branch="master"):
                return hashes_returned[branch]

            mocked_scm.return_value.get_latest = mocked_get_latest

            testmodule_variant_mmd_path = path.join(
                BASE_DIR, '..', 'staged_data', 'testmodule-variant.yaml')
            testmodule_variant_mmd = modulemd.ModuleMetadata()
            with open(testmodule_variant_mmd_path) as mmd_file:
                testmodule_variant_mmd.loads(mmd_file)

            module_build = \
                db.session.query(models.ModuleBuild).filter_by(id=1).one()
            mmd = module_build.mmd()

            error_msg = (
                'The included module "testmodule-variant" in "testmodule" have '
                'the following conflicting components: perl-List-Compare')
            try:
                module_build_service.utils.record_component_builds(
                    testmodule_variant_mmd, module_build, main_mmd=mmd)
                assert False, 'A RuntimeError was expected but was not raised'
            except RuntimeError as e:
                self.assertEqual(e.message, error_msg)

            self.assertEqual(module_build.state, models.BUILD_STATES['failed'])
            self.assertEqual(module_build.state_reason, error_msg)


class DummyModuleBuilder(GenericBuilder):
    """
    Dummy module builder
    """

    backend = "koji"
    _build_id = 0

    TAGGED_COMPONENTS = []

    @module_build_service.utils.validate_koji_tag('tag_name')
    def __init__(self, owner, module, config, tag_name, components):
        self.module_str = module
        self.tag_name = tag_name
        self.config = config

    def buildroot_connect(self, groups):
        pass

    def buildroot_prep(self):
        pass

    def buildroot_resume(self):
        pass

    def buildroot_ready(self, artifacts=None):
        return True

    def buildroot_add_dependency(self, dependencies):
        pass

    def buildroot_add_artifacts(self, artifacts, install=False):
        DummyModuleBuilder.TAGGED_COMPONENTS += artifacts

    def buildroot_add_repos(self, dependencies):
        pass

    def tag_artifacts(self, artifacts):
        pass

    @property
    def module_build_tag(self):
        return {"name": self.tag_name + "-build"}

    def build(self, artifact_name, source):
        DummyModuleBuilder._build_id += 1
        state = koji.BUILD_STATES['COMPLETE']
        reason = "Submitted %s to Koji" % (artifact_name)
        return DummyModuleBuilder._build_id, state, reason, None

    @staticmethod
    def get_disttag_srpm(disttag, module_build):
        # @FIXME
        return KojiModuleBuilder.get_disttag_srpm(disttag, module_build)

    def cancel_build(self, task_id):
        pass

    def list_tasks_for_components(self, component_builds=None, state='active'):
        pass


@patch("module_build_service.builder.GenericBuilder.default_buildroot_groups",
       return_value={'build': [], 'srpm-build': []})
class TestBatches(unittest.TestCase):

    def setUp(self):
        test_reuse_component_init_data()
        GenericBuilder.register_backend_class(DummyModuleBuilder)

    def tearDown(self):
        init_data()
        DummyModuleBuilder.TAGGED_COMPONENTS = []
        GenericBuilder.register_backend_class(KojiModuleBuilder)

    def test_start_next_batch_build_reuse(self, default_buildroot_groups):
        """
        Tests that start_next_batch_build:
           1) Increments module.batch.
           2) Can reuse all components in batch
           3) Returns proper further_work messages for reused components.
           4) Returns the fake Repo change message
           5) Handling the further_work messages lead to proper tagging of
              reused components.
        """
        module_build = models.ModuleBuild.query.filter_by(id=2).one()
        module_build.batch = 1

        builder = mock.MagicMock()
        further_work = module_build_service.utils.start_next_batch_build(
            conf, module_build, db.session, builder)

        # Batch number should increase.
        self.assertEqual(module_build.batch, 2)

        # KojiBuildChange messages in further_work should have build_new_state
        # set to COMPLETE, but the current component build state should be set
        # to BUILDING, so KojiBuildChange message handler handles the change
        # properly.
        for msg in further_work:
            if type(msg) == module_build_service.messaging.KojiBuildChange:
                self.assertEqual(msg.build_new_state, koji.BUILD_STATES['COMPLETE'])
                component_build = models.ComponentBuild.from_component_event(db.session, msg)
                self.assertEqual(component_build.state, koji.BUILD_STATES['BUILDING'])

        # When we handle these KojiBuildChange messages, MBS should tag all
        # the components just once.
        for msg in further_work:
            if type(msg) == module_build_service.messaging.KojiBuildChange:
                module_build_service.scheduler.handlers.components.complete(
                    conf, db.session, msg)

        # Since we have reused all the components in the batch, there should
        # be fake KojiRepoChange message.
        self.assertEqual(type(further_work[-1]), module_build_service.messaging.KojiRepoChange)

        # Check that packages have been tagged just once.
        self.assertEqual(len(DummyModuleBuilder.TAGGED_COMPONENTS), 2)

    @patch('module_build_service.utils.start_build_component')
    def test_start_next_batch_build_reuse_some(self, mock_sbc, default_buildroot_groups):
        """
        Tests that start_next_batch_build:
           1) Increments module.batch.
           2) Can reuse all components in the batch that it can.
           3) Returns proper further_work messages for reused components.
           4) Builds the remaining components
           5) Handling the further_work messages lead to proper tagging of
              reused components.
        """
        module_build = models.ModuleBuild.query.filter_by(id=2).one()
        module_build.batch = 1
        plc_component = models.ComponentBuild.query.filter_by(
            module_id=2, package='perl-List-Compare').one()
        plc_component.ref = '5ceea46add2366d8b8c5a623a2fb563b625b9abd'

        builder = mock.MagicMock()
        further_work = module_build_service.utils.start_next_batch_build(
            conf, module_build, db.session, builder)

        # Batch number should increase.
        self.assertEqual(module_build.batch, 2)

        # Make sure we only have one message returned for the one reused component
        self.assertEqual(len(further_work), 1)
        # The KojiBuildChange message in further_work should have build_new_state
        # set to COMPLETE, but the current component build state in the DB should be set
        # to BUILDING, so KojiBuildChange message handler handles the change
        # properly.
        self.assertEqual(further_work[0].build_new_state, koji.BUILD_STATES['COMPLETE'])
        component_build = models.ComponentBuild.from_component_event(db.session, further_work[0])
        self.assertEqual(component_build.state, koji.BUILD_STATES['BUILDING'])
        self.assertEqual(component_build.package, 'perl-Tangerine')
        self.assertIsNotNone(component_build.reused_component_id)
        # Make sure perl-List-Compare is set to the build state as well but not reused
        self.assertEqual(plc_component.state, koji.BUILD_STATES['BUILDING'])
        self.assertIsNone(plc_component.reused_component_id)
        mock_sbc.assert_called_once()

    @patch('module_build_service.utils.start_build_component')
    @patch('module_build_service.config.Config.rebuild_strategy',
           new_callable=mock.PropertyMock, return_value='all')
    def test_start_next_batch_build_rebuild_strategy_all(
            self, mock_rm, mock_sbc, default_buildroot_groups):
        """
        Tests that start_next_batch_build can't reuse any components in the batch because the
        rebuild method is set to "all".
        """
        module_build = models.ModuleBuild.query.filter_by(id=2).one()
        module_build.rebuild_strategy = 'all'
        module_build.batch = 1

        builder = mock.MagicMock()
        further_work = module_build_service.utils.start_next_batch_build(
            conf, module_build, db.session, builder)

        # Batch number should increase.
        self.assertEqual(module_build.batch, 2)
        # No component reuse messages should be returned
        self.assertEqual(len(further_work), 0)
        # Make sure that both components in the batch were submitted
        self.assertEqual(len(mock_sbc.mock_calls), 2)

    @patch('module_build_service.utils.start_build_component')
    @patch('module_build_service.config.Config.rebuild_strategy',
           new_callable=mock.PropertyMock, return_value='only-changed')
    def test_start_next_batch_build_rebuild_strategy_only_changed(
            self, mock_rm, mock_sbc, default_buildroot_groups):
        """
        Tests that start_next_batch_build reuses all unchanged components in the batch because the
        rebuild method is set to "only-changed". This means that one component is reused in batch
        2, and even though the other component in batch 2 changed and was rebuilt, the component
        in batch 3 can be reused.
        """
        module_build = models.ModuleBuild.query.filter_by(id=2).one()
        module_build.rebuild_strategy = 'only-changed'
        module_build.batch = 1
        # perl-List-Compare changed
        plc_component = models.ComponentBuild.query.filter_by(
            module_id=2, package='perl-List-Compare').one()
        plc_component.ref = '5ceea46add2366d8b8c5a623a2fb563b625b9abd'

        builder = mock.MagicMock()
        further_work = module_build_service.utils.start_next_batch_build(
            conf, module_build, db.session, builder)

        # Batch number should increase
        self.assertEqual(module_build.batch, 2)

        # Make sure we only have one message returned for the one reused component
        self.assertEqual(len(further_work), 1)
        # The KojiBuildChange message in further_work should have build_new_state
        # set to COMPLETE, but the current component build state in the DB should be set
        # to BUILDING, so KojiBuildChange message handler handles the change
        # properly.
        self.assertEqual(further_work[0].build_new_state, koji.BUILD_STATES['COMPLETE'])
        component_build = models.ComponentBuild.from_component_event(db.session, further_work[0])
        self.assertEqual(component_build.state, koji.BUILD_STATES['BUILDING'])
        self.assertEqual(component_build.package, 'perl-Tangerine')
        self.assertIsNotNone(component_build.reused_component_id)
        # Make sure perl-List-Compare is set to the build state as well but not reused
        self.assertEqual(plc_component.state, koji.BUILD_STATES['BUILDING'])
        self.assertIsNone(plc_component.reused_component_id)
        mock_sbc.assert_called_once()
        mock_sbc.reset_mock()

        # Complete the build
        plc_component.state = koji.BUILD_STATES['COMPLETE']
        pt_component = models.ComponentBuild.query.filter_by(
            module_id=2, package='perl-Tangerine').one()
        pt_component.state = koji.BUILD_STATES['COMPLETE']

        # Start the next build batch
        further_work = module_build_service.utils.start_next_batch_build(
            conf, module_build, db.session, builder)
        # Batch number should increase
        self.assertEqual(module_build.batch, 3)
        # Verify that tangerine was reused even though perl-Tangerine was rebuilt in the previous
        # batch
        self.assertEqual(further_work[0].build_new_state, koji.BUILD_STATES['COMPLETE'])
        component_build = models.ComponentBuild.from_component_event(db.session, further_work[0])
        self.assertEqual(component_build.state, koji.BUILD_STATES['BUILDING'])
        self.assertEqual(component_build.package, 'tangerine')
        self.assertIsNotNone(component_build.reused_component_id)
        mock_sbc.assert_not_called()

    @patch('module_build_service.utils.start_build_component')
    def test_start_next_batch_build_smart_scheduling(self, mock_sbc, default_buildroot_groups):
        """
        Tests that components with the longest build time will be scheduled first
        """
        module_build = models.ModuleBuild.query.filter_by(id=2).one()
        module_build.batch = 1
        pt_component = models.ComponentBuild.query.filter_by(
            module_id=2, package='perl-Tangerine').one()
        pt_component.ref = '6ceea46add2366d8b8c5a623b2fb563b625bfabe'
        plc_component = models.ComponentBuild.query.filter_by(
            module_id=2, package='perl-List-Compare').one()
        plc_component.ref = '5ceea46add2366d8b8c5a623a2fb563b625b9abd'

        builder = mock.MagicMock()
        # The call order of get_average_build_time should be by the component's ID. Having this
        # side_effect tells continue_batch_build to build the second component in the build batch
        # first and the first component in the build batch second.
        builder.get_average_build_time.side_effect = [1234.56, 2345.67]
        further_work = module_build_service.utils.start_next_batch_build(
            conf, module_build, db.session, builder)

        # Batch number should increase.
        self.assertEqual(module_build.batch, 2)

        # Make sure we don't have any messages returned since no components should be reused
        self.assertEqual(len(further_work), 0)
        # Make sure both components are set to the build state but not reused
        self.assertEqual(pt_component.state, koji.BUILD_STATES['BUILDING'])
        self.assertIsNone(pt_component.reused_component_id)
        self.assertEqual(plc_component.state, koji.BUILD_STATES['BUILDING'])
        self.assertIsNone(plc_component.reused_component_id)
        # Test the order of the scheduling
        expected_calls = [mock.call(builder, plc_component), mock.call(builder, pt_component)]
        self.assertEqual(mock_sbc.mock_calls, expected_calls)

    @patch('module_build_service.utils.start_build_component')
    def test_start_next_batch_continue(self, mock_sbc, default_buildroot_groups):
        """
        Tests that start_next_batch_build does not start new batch when
        there are unbuilt components in the current one.
        """
        module_build = models.ModuleBuild.query.filter_by(id=2).one()
        module_build.batch = 2

        # The component was reused when the batch first started
        building_component = module_build.current_batch()[0]
        building_component.state = koji.BUILD_STATES['BUILDING']
        building_component.reused_component_id = 123
        db.session.commit()

        builder = mock.MagicMock()
        further_work = module_build_service.utils.start_next_batch_build(
            conf, module_build, db.session, builder)

        # Batch number should not increase.
        self.assertEqual(module_build.batch, 2)
        # Make sure start build was called for the second component which wasn't reused
        mock_sbc.assert_called_once()
        # No further work should be returned
        self.assertEqual(len(further_work), 0)

    def test_start_next_batch_build_repo_building(self, default_buildroot_groups):
        """
        Test that start_next_batch_build does not start new batch when
        builder.buildroot_ready() returns False.
        """
        module_build = models.ModuleBuild.query.filter_by(id=2).one()
        module_build.batch = 1

        builder = mock.MagicMock()
        builder.buildroot_ready.return_value = False

        # Batch number should not increase.
        self.assertEqual(module_build.batch, 1)


@patch("module_build_service.config.Config.mock_resultsdir",
       new_callable=mock.PropertyMock,
       return_value=path.join(
           BASE_DIR, '..', 'staged_data', "local_builds"))
@patch("module_build_service.config.Config.system",
       new_callable=mock.PropertyMock, return_value="mock")
class TestLocalBuilds(unittest.TestCase):

    def setUp(self):
        init_data()

    def tearDown(self):
        init_data()

    def test_load_local_builds_name(self, conf_system, conf_resultsdir):
        with app.app_context():
            module_build_service.utils.load_local_builds("testmodule")
            local_modules = models.ModuleBuild.local_modules(db.session)

            self.assertEqual(len(local_modules), 1)
            self.assertTrue(local_modules[0].koji_tag.endswith(
                "/module-testmodule-master-20170816080816/results"))

    def test_load_local_builds_name_stream(
            self, conf_system, conf_resultsdir):
        with app.app_context():
            module_build_service.utils.load_local_builds("testmodule:master")
            local_modules = models.ModuleBuild.local_modules(db.session)

            self.assertEqual(len(local_modules), 1)
            self.assertTrue(local_modules[0].koji_tag.endswith(
                "/module-testmodule-master-20170816080816/results"))

    def test_load_local_builds_name_stream_non_existing(
            self, conf_system, conf_resultsdir):
        with app.app_context():
            with self.assertRaises(RuntimeError):
                module_build_service.utils.load_local_builds("testmodule:x")
                models.ModuleBuild.local_modules(db.session)

    def test_load_local_builds_name_stream_version(
            self, conf_system, conf_resultsdir):
        with app.app_context():
            module_build_service.utils.load_local_builds("testmodule:master:20170816080815")
            local_modules = models.ModuleBuild.local_modules(db.session)

            self.assertEqual(len(local_modules), 1)
            self.assertTrue(local_modules[0].koji_tag.endswith(
                "/module-testmodule-master-20170816080815/results"))

    def test_load_local_builds_name_stream_version_non_existing(
            self, conf_system, conf_resultsdir):
        with app.app_context():
            with self.assertRaises(RuntimeError):
                module_build_service.utils.load_local_builds("testmodule:master:123")
                models.ModuleBuild.local_modules(db.session)

    def test_load_local_builds_base_runtime(
            self, conf_system, conf_resultsdir):
        with app.app_context():
            module_build_service.utils.load_local_builds("base-runtime")
            local_modules = models.ModuleBuild.local_modules(db.session)

            self.assertEqual(len(local_modules), 1)
            self.assertTrue(local_modules[0].koji_tag.endswith(
                "/module-base-runtime-master-20170816080815/results"))

    def test_load_local_builds_base_runtime_master(
            self, conf_system, conf_resultsdir):
        with app.app_context():
            module_build_service.utils.load_local_builds("base-runtime:master")
            local_modules = models.ModuleBuild.local_modules(db.session)

            self.assertEqual(len(local_modules), 1)
            self.assertTrue(local_modules[0].koji_tag.endswith(
                "/module-base-runtime-master-20170816080815/results"))



# Copyright (c) 2016  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Written by Matt Prahl <mprahl@redhat.com

import unittest
import json
import vcr

import modulemd as _modulemd
import module_build_service.scm

from mock import patch, PropertyMock
from shutil import copyfile
from os import path, mkdir
from os.path import dirname
import hashlib

from tests import app, init_data
from module_build_service.errors import UnprocessableEntity
from module_build_service.models import ModuleBuild
from module_build_service import db, version
import module_build_service.config as mbs_config
import module_build_service.scheduler.handlers.modules


user = ('Homer J. Simpson', set(['packager']))
other_user = ('some_other_user', set(['packager']))
anonymous_user = ('anonymous', set(['packager']))
base_dir = dirname(dirname(__file__))
cassette_dir = base_dir + '/vcr-request-data/'


class FakeSCM(object):
    def __init__(self, mocked_scm, name, mmd_filenames, commit=None, checkout_raise=False,
                 get_latest_raise=False):
        """
        Adds default testing checkout, get_latest and name methods
        to mocked_scm SCM class.

        :param mmd_filenames: List of ModuleMetadata yaml files which
        will be checkouted by the SCM class in the same order as they
        are stored in the list.
        """
        self.mocked_scm = mocked_scm
        self.name = name
        self.commit = commit
        if not isinstance(mmd_filenames, list):
            mmd_filenames = [mmd_filenames]
        self.mmd_filenames = mmd_filenames
        self.checkout_id = 0
        self.sourcedir = None

        if checkout_raise:
            self.mocked_scm.return_value.checkout.side_effect = \
                UnprocessableEntity(
                    "checkout: The requested commit hash was not found within "
                    "the repository. Perhaps you forgot to push. The original "
                    "message was: ")
        else:
            self.mocked_scm.return_value.checkout = self.checkout

        self.mocked_scm.return_value.name = self.name
        self.mocked_scm.return_value.commit = self.commit
        if get_latest_raise:
            self.mocked_scm.return_value.get_latest.side_effect = \
                RuntimeError("Failed to get_latest commit")
        else:
            self.mocked_scm.return_value.get_latest = self.get_latest
        self.mocked_scm.return_value.repository_root = "git://pkgs.stg.fedoraproject.org/modules/"
        self.mocked_scm.return_value.branch = 'master'
        self.mocked_scm.return_value.sourcedir = self.sourcedir
        self.mocked_scm.return_value.get_module_yaml = self.get_module_yaml

    def checkout(self, temp_dir):
        try:
            mmd_filename = self.mmd_filenames[self.checkout_id]
        except Exception:
            mmd_filename = self.mmd_filenames[0]

        self.sourcedir = path.join(temp_dir, self.name)
        mkdir(self.sourcedir)
        base_dir = path.abspath(path.dirname(__file__))
        copyfile(path.join(base_dir, '..', 'staged_data', mmd_filename),
                 self.get_module_yaml())

        self.checkout_id += 1

        return self.sourcedir

    def get_latest(self, branch='master'):
        return hashlib.sha1(branch).hexdigest()[:10]

    def get_module_yaml(self):
        return path.join(self.sourcedir, self.name + ".yaml")


class TestViews(unittest.TestCase):
    maxDiff = None

    def setUp(self):
        self.client = app.test_client()
        init_data()

        filename = cassette_dir + self.id()
        self.vcr = vcr.use_cassette(filename)
        self.vcr.__enter__()

    def tearDown(self):
        self.vcr.__exit__()

    def test_query_build(self):
        rv = self.client.get('/module-build-service/1/module-builds/1')
        data = json.loads(rv.data)
        self.assertEquals(data['id'], 1)
        self.assertEquals(data['name'], 'nginx')
        self.assertEquals(data['owner'], 'Moe Szyslak')
        self.assertEquals(data['state'], 3)
        self.assertEquals(data['state_reason'], None)
        self.assertDictEqual(data['tasks'], {
            'rpms': {
                'module-build-macros': {
                    'task_id': 12312321,
                    'state': 1,
                    'state_reason': None,
                    'nvr': 'module-build-macros-01-1.module_nginx_1_2',
                },
                'nginx': {
                    'task_id': 12312345,
                    'state': 1,
                    'state_reason': None,
                    'nvr': 'nginx-1.10.1-2.module_nginx_1_2',
                },
            },
        })
        self.assertEquals(data['time_completed'], '2016-09-03T11:25:32Z')
        self.assertEquals(data['time_modified'], '2016-09-03T11:25:32Z')
        self.assertEquals(data['time_submitted'], '2016-09-03T11:23:20Z')
        self.assertEqual(data['rebuild_strategy'], 'changed-and-after')

    def test_query_build_with_verbose_mode(self):
        rv = self.client.get('/module-build-service/1/module-builds/1?verbose=true')
        data = json.loads(rv.data)
        self.assertEquals(data['component_builds'], [1, 2])
        self.assertEquals(data['id'], 1)
        with open(path.join(base_dir, "staged_data", "nginx_mmd.yaml")) as mmd:
            self.assertEquals(data['modulemd'], mmd.read())
        self.assertEquals(data['name'], 'nginx')
        self.assertEquals(data['owner'], 'Moe Szyslak')
        self.assertEquals(data['scmurl'],
                          ('git://pkgs.domain.local/modules/nginx'
                           '?#ba95886c7a443b36a9ce31abda1f9bef22f2f8c9'))
        self.assertEquals(data['state'], 3)
        self.assertEquals(data['state_name'], 'done')
        self.assertEquals(data['state_reason'], None)
        self.assertEquals(data['state_trace'][0]['reason'], None)
        self.assertTrue(data['state_trace'][0]['time'] is not None)
        self.assertEquals(data['state_trace'][0]['state'], 3)
        self.assertEquals(data['state_trace'][0]['state_name'], 'done')
        self.assertEquals(data['state_url'], '/module-build-service/1/module-builds/1')
        self.assertEquals(data['stream'], '1')
        self.assertDictEqual(data['tasks'], {
            'rpms': {
                'module-build-macros': {
                    'task_id': 12312321,
                    'state': 1,
                    'state_reason': None,
                    'nvr': 'module-build-macros-01-1.module_nginx_1_2',
                },
                'nginx': {
                    'task_id': 12312345,
                    'state': 1,
                    'state_reason': None,
                    'nvr': 'nginx-1.10.1-2.module_nginx_1_2',
                },
            },
        })
        self.assertEquals(data['time_completed'], u'2016-09-03T11:25:32Z')
        self.assertEquals(data['time_modified'], u'2016-09-03T11:25:32Z')
        self.assertEquals(data['time_submitted'], u'2016-09-03T11:23:20Z')
        self.assertEquals(data['version'], '2')
        self.assertEqual(data['rebuild_strategy'], 'changed-and-after')

    def test_pagination_metadata(self):
        rv = self.client.get('/module-build-service/1/module-builds/?per_page=8&page=2')
        meta_data = json.loads(rv.data)['meta']
        self.assertIn(
            meta_data['prev'].split('?', 1)[1], ['per_page=8&page=1', 'page=1&per_page=8'])
        self.assertIn(
            meta_data['next'].split('?', 1)[1], ['per_page=8&page=3', 'page=3&per_page=8'])
        self.assertIn(
            meta_data['last'].split('?', 1)[1], ['per_page=8&page=4', 'page=4&per_page=8'])
        self.assertIn(
            meta_data['first'].split('?', 1)[1], ['per_page=8&page=1', 'page=1&per_page=8'])
        self.assertEquals(meta_data['total'], 30)
        self.assertEquals(meta_data['per_page'], 8)
        self.assertEquals(meta_data['pages'], 4)
        self.assertEquals(meta_data['page'], 2)

    def test_pagination_metadata_with_args(self):
        rv = self.client.get('/module-build-service/1/module-builds/?per_page=8&page=2&order_by=id')
        meta_data = json.loads(rv.data)['meta']
        for link in [meta_data['prev'], meta_data['next'], meta_data['last'], meta_data['first']]:
            self.assertIn('order_by=id', link)
            self.assertIn('per_page=8', link)
        self.assertEquals(meta_data['total'], 30)
        self.assertEquals(meta_data['per_page'], 8)
        self.assertEquals(meta_data['pages'], 4)
        self.assertEquals(meta_data['page'], 2)

    def test_query_builds(self):
        rv = self.client.get('/module-build-service/1/module-builds/?per_page=2')
        items = json.loads(rv.data)['items']
        expected = [
            {
                'id': 30,
                'koji_tag': None,
                'name': 'testmodule',
                'rebuild_strategy': 'changed-and-after',
                'owner': 'some_other_user',
                'scmurl': ('git://pkgs.domain.local/modules/testmodule?'
                           '#ca95886c7a443b36a9ce31abda1f9bef22f2f8c9'),
                'state': 1,
                'state_name': 'wait',
                'state_reason': None,
                'stream': '4.3.43',
                'tasks': {
                    'rpms': {
                        'module-build-macros': {
                            'nvr': 'module-build-macros-01-1.module_postgresql_1_2',
                            'state': 1,
                            'state_reason': None,
                            'task_id': 47384002
                        },
                        'rubygem-rails': {
                            'nvr': 'postgresql-9.5.3-4.module_postgresql_1_2',
                            'state': 3,
                            'state_reason': None,
                            'task_id': 2433442
                        }
                    }
                },
                'time_completed': None,
                'time_modified': '2016-09-03T13:58:40Z',
                'time_submitted': '2016-09-03T13:58:33Z',
                'version': '6'
            },
            {
                'id': 29,
                'koji_tag': 'module-postgressql-1.2',
                'name': 'postgressql',
                'owner': 'some_user',
                'rebuild_strategy': 'changed-and-after',
                'scmurl': ('git://pkgs.domain.local/modules/postgressql?'
                           '#aa95886c7a443b36a9ce31abda1f9bef22f2f8c9'),
                'state': 3,
                'state_name': 'done',
                'state_reason': None,
                'stream': '1',
                'tasks': {
                    'rpms': {
                        'module-build-macros': {
                            'nvr': 'module-build-macros-01-1.module_postgresql_1_2',
                            'state': 1,
                            'state_reason': None,
                            'task_id': 47384002
                        },
                        'postgresql': {
                            'nvr': 'postgresql-9.5.3-4.module_postgresql_1_2',
                            'state': 1,
                            'state_reason': None,
                            'task_id': 2433442
                        }
                    }
                },
                'time_completed': '2016-09-03T12:57:19Z',
                'time_modified': '2016-09-03T13:57:19Z',
                'time_submitted': '2016-09-03T13:55:33Z',
                'version': '2'
            }
        ]
        self.assertEquals(items, expected)

    def test_query_component_build(self):
        rv = self.client.get('/module-build-service/1/component-builds/1')
        data = json.loads(rv.data)
        self.assertEquals(data['id'], 1)
        self.assertEquals(data['format'], 'rpms')
        self.assertEquals(data['module_build'], 1)
        self.assertEquals(data['package'], 'nginx')
        self.assertEquals(data['state'], 1)
        self.assertEquals(data['state_name'], 'COMPLETE')
        self.assertEquals(data['state_reason'], None)
        self.assertEquals(data['task_id'], 12312345)

    def test_query_component_build_verbose(self):
        rv = self.client.get('/module-build-service/1/component-builds/3?verbose=true')
        data = json.loads(rv.data)
        self.assertEquals(data['id'], 3)
        self.assertEquals(data['format'], 'rpms')
        self.assertEquals(data['module_build'], 2)
        self.assertEquals(data['package'], 'postgresql')
        self.assertEquals(data['state'], 1)
        self.assertEquals(data['state_name'], 'COMPLETE')
        self.assertEquals(data['state_reason'], None)
        self.assertEquals(data['task_id'], 2433433)
        self.assertEquals(data['state_trace'][0]['reason'], None)
        self.assertTrue(data['state_trace'][0]['time'] is not None)
        self.assertEquals(data['state_trace'][0]['state'], 1)
        self.assertEquals(data['state_trace'][0]['state_name'], 'wait')

    component_builds_filters = ['tagged', 'ref', 'format']

    def test_query_component_builds_filter_format(self):
        rv = self.client.get('/module-build-service/1/component-builds/'
                             '?format=rpms')
        data = json.loads(rv.data)
        self.assertEquals(data['meta']['total'], 60)

    def test_query_component_builds_filter_ref(self):
        rv = self.client.get('/module-build-service/1/component-builds/'
                             '?ref=this-filter-query-should-return-zero-items')
        data = json.loads(rv.data)
        self.assertEquals(data['meta']['total'], 0)

    def test_query_component_builds_filter_tagged(self):
        rv = self.client.get('/module-build-service/1/component-builds/'
                             '?tagged=this-filter-query-should-return-zero-items')
        data = json.loads(rv.data)
        self.assertEquals(data['meta']['total'], 0)

    def test_query_component_builds_filter_nvr(self):
        rv = self.client.get(
            '/module-build-service/1/component-builds/?nvr=nginx-1.10.1-2.module_nginx_1_2'
        )
        data = json.loads(rv.data)
        self.assertEquals(data['meta']['total'], 10)

    def test_query_component_builds_filter_task_id(self):
        rv = self.client.get('/module-build-service/1/component-builds/?task_id=12312346')
        data = json.loads(rv.data)
        self.assertEquals(data['meta']['total'], 1)

    def test_query_builds_filter_name(self):
        rv = self.client.get('/module-build-service/1/module-builds/?name=nginx')
        data = json.loads(rv.data)
        self.assertEquals(data['meta']['total'], 10)

    def test_query_builds_filter_koji_tag(self):
        rv = self.client.get('/module-build-service/1/module-builds/?koji_tag=module-nginx-1.2')
        data = json.loads(rv.data)
        self.assertEquals(data['meta']['total'], 10)

    def test_query_builds_filter_completed_before(self):
        rv = self.client.get(
            '/module-build-service/1/module-builds/?completed_before=2016-09-03T11:30:00Z')
        data = json.loads(rv.data)
        self.assertEquals(data['meta']['total'], 2)

    def test_query_builds_filter_completed_after(self):
        rv = self.client.get(
            '/module-build-service/1/module-builds/?completed_after=2016-09-03T12:25:00Z')
        data = json.loads(rv.data)
        self.assertEquals(data['meta']['total'], 8)

    def test_query_builds_filter_submitted_before(self):
        rv = self.client.get(
            '/module-build-service/1/module-builds/?submitted_before=2016-09-03T12:25:00Z')
        data = json.loads(rv.data)
        self.assertEquals(data['meta']['total'], 7)

    def test_query_builds_filter_submitted_after(self):
        rv = self.client.get(
            '/module-build-service/1/module-builds/?submitted_after=2016-09-03T12:25:00Z')
        data = json.loads(rv.data)
        self.assertEquals(data['meta']['total'], 23)

    def test_query_builds_filter_modified_before(self):
        rv = self.client.get(
            '/module-build-service/1/module-builds/?modified_before=2016-09-03T12:25:00Z')
        data = json.loads(rv.data)
        self.assertEquals(data['meta']['total'], 6)

    def test_query_builds_filter_modified_after(self):
        rv = self.client.get(
            '/module-build-service/1/module-builds/?modified_after=2016-09-03T12:25:00Z')
        data = json.loads(rv.data)
        self.assertEquals(data['meta']['total'], 24)

    def test_query_builds_filter_owner(self):
        rv = self.client.get(
            '/module-build-service/1/module-builds/?owner=Moe%20Szyslak')
        data = json.loads(rv.data)
        self.assertEquals(data['meta']['total'], 10)

    def test_query_builds_filter_state(self):
        rv = self.client.get(
            '/module-build-service/1/module-builds/?state=3')
        data = json.loads(rv.data)
        self.assertEquals(data['meta']['total'], 20)

    def test_query_builds_two_filters(self):
        rv = self.client.get('/module-build-service/1/module-builds/?owner=Moe%20Szyslak'
                             '&modified_after=2016-09-03T12:25:00Z')
        data = json.loads(rv.data)
        self.assertEquals(data['meta']['total'], 4)

    def test_query_builds_filter_invalid_date(self):
        rv = self.client.get(
            '/module-build-service/1/module-builds/?modified_after=2016-09-03T12:25:00-05:00')
        data = json.loads(rv.data)
        self.assertEquals(data['error'], 'Bad Request')
        self.assertEquals(data['message'], 'An invalid Zulu ISO 8601 timestamp'
                          ' was provided for the \"modified_after\" parameter')
        self.assertEquals(data['status'], 400)

    def test_query_builds_order_by(self):
        build = db.session.query(module_build_service.models.ModuleBuild).filter_by(id=2).one()
        build.name = 'candy'
        db.session.add(build)
        db.session.commit()
        rv = self.client.get('/module-build-service/1/module-builds/?'
                             'per_page=10&order_by=name')
        items = json.loads(rv.data)['items']
        self.assertEqual(items[0]['name'], 'candy')
        self.assertEqual(items[1]['name'], 'nginx')

    def test_query_builds_order_desc_by(self):
        rv = self.client.get('/module-build-service/1/module-builds/?'
                             'per_page=10&order_desc_by=id')
        items = json.loads(rv.data)['items']
        # Check that the id is items[0]["id"], items[0]["id"] - 1, ...
        for idx, item in enumerate(items):
            self.assertEquals(item["id"], items[0]["id"] - idx)

    def test_query_builds_order_by_order_desc_by(self):
        """
        Test that when both order_by and order_desc_by is set,
        we prefer order_desc_by.
        """
        rv = self.client.get('/module-build-service/1/module-builds/?'
                             'per_page=10&order_desc_by=id&order_by=name')
        items = json.loads(rv.data)['items']
        # Check that the id is items[0]["id"], items[0]["id"] - 1, ...
        for idx, item in enumerate(items):
            self.assertEquals(item["id"], items[0]["id"] - idx)

    def test_query_builds_order_by_wrong_key(self):
        rv = self.client.get('/module-build-service/1/module-builds/?'
                             'per_page=10&order_by=unknown')
        data = json.loads(rv.data)
        self.assertEquals(data['status'], 400)
        self.assertEquals(data['error'], 'Bad Request')
        self.assertEquals(
            data['message'], 'An invalid order_by or order_desc_by key '
            'was supplied')

    @patch('module_build_service.auth.get_user', return_value=user)
    @patch('module_build_service.scm.SCM')
    def test_submit_build(self, mocked_scm, mocked_get_user):
        FakeSCM(mocked_scm, 'testmodule', 'testmodule.yaml',
                '620ec77321b2ea7b0d67d82992dda3e1d67055b4')

        rv = self.client.post('/module-build-service/1/module-builds/', data=json.dumps(
            {'branch': 'master', 'scmurl': 'git://pkgs.stg.fedoraproject.org/modules/'
                'testmodule.git?#68931c90de214d9d13feefbd35246a81b6cb8d49'}))
        data = json.loads(rv.data)

        assert 'component_builds' in data, data
        self.assertEquals(data['component_builds'], [])
        self.assertEquals(data['name'], 'testmodule')
        self.assertEquals(data['scmurl'],
                          ('git://pkgs.stg.fedoraproject.org/modules/testmodule'
                          '.git?#68931c90de214d9d13feefbd35246a81b6cb8d49'))
        self.assertEquals(data['version'], '1')
        self.assertTrue(data['time_submitted'] is not None)
        self.assertTrue(data['time_modified'] is not None)
        self.assertEquals(data['time_completed'], None)
        self.assertEquals(data['stream'], 'master')
        self.assertEquals(data['owner'], 'Homer J. Simpson')
        self.assertEquals(data['id'], 31)
        self.assertEquals(data['rebuild_strategy'], 'changed-and-after')
        self.assertEquals(data['state_name'], 'init')
        self.assertEquals(data['state_url'], '/module-build-service/1/module-builds/31')
        self.assertEquals(data['state_trace'], [])
        self.assertDictEqual(data['tasks'], {})
        mmd = _modulemd.ModuleMetadata()
        mmd.loads(data["modulemd"])

    @patch('module_build_service.auth.get_user', return_value=user)
    @patch('module_build_service.scm.SCM')
    @patch('module_build_service.config.Config.rebuild_strategy_allow_override',
           new_callable=PropertyMock, return_value=True)
    def test_submit_build_rebuild_strategy(self, mocked_rmao, mocked_scm, mocked_get_user):
        FakeSCM(mocked_scm, 'testmodule', 'testmodule.yaml',
                '620ec77321b2ea7b0d67d82992dda3e1d67055b4')

        rv = self.client.post('/module-build-service/1/module-builds/', data=json.dumps(
            {'branch': 'master', 'rebuild_strategy': 'only-changed',
             'scmurl': ('git://pkgs.stg.fedoraproject.org/modules/testmodule.git?'
                        '#68931c90de214d9d13feefbd35246a81b6cb8d49')}))
        data = json.loads(rv.data)
        self.assertEquals(data['rebuild_strategy'], 'only-changed')

    @patch('module_build_service.auth.get_user', return_value=user)
    @patch('module_build_service.scm.SCM')
    @patch('module_build_service.config.Config.rebuild_strategies_allowed',
           new_callable=PropertyMock, return_value=['all'])
    @patch('module_build_service.config.Config.rebuild_strategy_allow_override',
           new_callable=PropertyMock, return_value=True)
    def test_submit_build_rebuild_strategy_not_allowed(self, mock_rsao, mock_rsa, mocked_scm,
                                                       mocked_get_user):
        FakeSCM(mocked_scm, 'testmodule', 'testmodule.yaml',
                '620ec77321b2ea7b0d67d82992dda3e1d67055b4')

        rv = self.client.post('/module-build-service/1/module-builds/', data=json.dumps(
            {'branch': 'master', 'rebuild_strategy': 'only-changed',
                'scmurl': ('git://pkgs.stg.fedoraproject.org/modules/testmodule.git?'
                        '#68931c90de214d9d13feefbd35246a81b6cb8d49')}))
        data = json.loads(rv.data)
        self.assertEqual(rv.status_code, 400)
        expected_error = {
            'error': 'Bad Request',
            'message': ('The rebuild method of "only-changed" is not allowed. Choose from: all.'),
            'status': 400
        }
        self.assertEqual(data, expected_error)

    @patch('module_build_service.auth.get_user', return_value=user)
    @patch('module_build_service.scm.SCM')
    def test_submit_build_rebuild_strategy_override_not_allowed(self, mocked_scm, mocked_get_user):
        FakeSCM(mocked_scm, 'testmodule', 'testmodule.yaml',
                '620ec77321b2ea7b0d67d82992dda3e1d67055b4')

        rv = self.client.post('/module-build-service/1/module-builds/', data=json.dumps(
            {'branch': 'master', 'rebuild_strategy': 'only-changed',
             'scmurl': ('git://pkgs.stg.fedoraproject.org/modules/testmodule.git?'
                        '#68931c90de214d9d13feefbd35246a81b6cb8d49')}))
        data = json.loads(rv.data)
        self.assertEqual(rv.status_code, 400)
        expected_error = {
            'error': 'Bad Request',
            'message': ('The request contains the "rebuild_strategy" parameter but overriding '
                        'the default isn\'t allowed'),
            'status': 400
        }
        self.assertEqual(data, expected_error)

    @patch('module_build_service.auth.get_user', return_value=user)
    @patch('module_build_service.scm.SCM')
    def test_submit_componentless_build(self, mocked_scm, mocked_get_user):
        FakeSCM(mocked_scm, 'fakemodule', 'fakemodule.yaml',
                '3da541559918a808c2402bba5012f6c60b27661c')

        rv = self.client.post('/module-build-service/1/module-builds/', data=json.dumps(
            {'branch': 'master', 'scmurl': 'git://pkgs.stg.fedoraproject.org/modules/'
                'testmodule.git?#68931c90de214d9d13feefbd35246a81b6cb8d49'}))
        data = json.loads(rv.data)

        self.assertEquals(data['component_builds'], [])
        self.assertEquals(data['name'], 'fakemodule')
        self.assertEquals(data['scmurl'],
                          ('git://pkgs.stg.fedoraproject.org/modules/testmodule'
                          '.git?#68931c90de214d9d13feefbd35246a81b6cb8d49'))
        self.assertEquals(data['version'], '1')
        self.assertTrue(data['time_submitted'] is not None)
        self.assertTrue(data['time_modified'] is not None)
        self.assertEquals(data['time_completed'], None)
        self.assertEquals(data['stream'], 'master')
        self.assertEquals(data['owner'], 'Homer J. Simpson')
        self.assertEquals(data['id'], 31)
        self.assertEquals(data['state_name'], 'init')
        self.assertEquals(data['rebuild_strategy'], 'changed-and-after')

    def test_submit_build_auth_error(self):
        base_dir = path.abspath(path.dirname(__file__))
        client_secrets = path.join(base_dir, "client_secrets.json")
        with patch.dict('module_build_service.app.config', {'OIDC_CLIENT_SECRETS': client_secrets}):
            rv = self.client.post('/module-build-service/1/module-builds/', data=json.dumps(
                {'branch': 'master', 'scmurl': 'git://pkgs.stg.fedoraproject.org/modules/'
                    'testmodule.git?#48931b90de214d9d13feefbd35246a81b6cb8d49'}))
            data = json.loads(rv.data)
            self.assertEquals(
                data['message'],
                "No 'authorization' header found."
            )
            self.assertEquals(data['status'], 401)
            self.assertEquals(data['error'], 'Unauthorized')

    @patch('module_build_service.auth.get_user', return_value=user)
    def test_submit_build_scm_url_error(self, mocked_get_user):
        rv = self.client.post('/module-build-service/1/module-builds/', data=json.dumps(
            {'branch': 'master', 'scmurl': 'git://badurl.com'}))
        data = json.loads(rv.data)
        self.assertEquals(data['message'], 'The submitted scmurl '
                          'git://badurl.com is not allowed')
        self.assertEquals(data['status'], 403)
        self.assertEquals(data['error'], 'Forbidden')

    @patch('module_build_service.auth.get_user', return_value=user)
    def test_submit_build_scm_url_without_hash(self, mocked_get_user):
        rv = self.client.post('/module-build-service/1/module-builds/', data=json.dumps(
            {'branch': 'master', 'scmurl': 'git://pkgs.stg.fedoraproject.org/modules/'
                'testmodule.git'}))
        data = json.loads(rv.data)
        self.assertEquals(data['message'], 'The submitted scmurl '
                          'git://pkgs.stg.fedoraproject.org/modules/testmodule.git '
                          'is not valid')
        self.assertEquals(data['status'], 403)
        self.assertEquals(data['error'], 'Forbidden')

    @patch('module_build_service.auth.get_user', return_value=user)
    @patch('module_build_service.scm.SCM')
    def test_submit_build_bad_modulemd(self, mocked_scm, mocked_get_user):
        FakeSCM(mocked_scm, "bad", "bad.yaml")

        rv = self.client.post('/module-build-service/1/module-builds/', data=json.dumps(
            {'branch': 'master', 'scmurl': 'git://pkgs.stg.fedoraproject.org/modules/'
                'testmodule.git?#68931c90de214d9d13feefbd35246a81b6cb8d49'}))
        data = json.loads(rv.data)
        self.assertTrue(data['message'].startswith('Invalid modulemd:'))
        self.assertEquals(data['status'], 422)
        self.assertEquals(data['error'], 'Unprocessable Entity')

    @patch('module_build_service.auth.get_user', return_value=user)
    @patch('module_build_service.scm.SCM')
    def test_submit_build_includedmodule_custom_repo_not_allowed(self,
                                                                 mocked_scm, mocked_get_user):
        FakeSCM(mocked_scm, "includedmodules", ["includedmodules.yaml",
                                                "testmodule.yaml"])
        rv = self.client.post('/module-build-service/1/module-builds/', data=json.dumps(
            {'branch': 'master', 'scmurl': 'git://pkgs.stg.fedoraproject.org/modules/'
                'testmodule.git?#68931c90de214d9d13feefbd35246a81b6cb8d49'}))
        data = json.loads(rv.data)

        self.assertEquals(data['status'], 403)
        self.assertEquals(data['error'], 'Forbidden')

    @patch('module_build_service.auth.get_user', return_value=other_user)
    def test_cancel_build(self, mocked_get_user):
        rv = self.client.patch('/module-build-service/1/module-builds/30',
                               data=json.dumps({'state': 'failed'}))
        data = json.loads(rv.data)

        self.assertEquals(data['state'], 4)
        self.assertEquals(data['state_reason'], 'Canceled by some_other_user.')

    @patch('module_build_service.auth.get_user', return_value=('sammy', set()))
    def test_cancel_build_unauthorized_no_groups(self, mocked_get_user):
        rv = self.client.patch('/module-build-service/1/module-builds/30',
                               data=json.dumps({'state': 'failed'}))
        data = json.loads(rv.data)

        self.assertEquals(data['status'], 403)
        self.assertEquals(data['error'], 'Forbidden')

    @patch('module_build_service.auth.get_user', return_value=('sammy', set(["packager"])))
    def test_cancel_build_unauthorized_not_owner(self, mocked_get_user):
        rv = self.client.patch('/module-build-service/1/module-builds/30',
                               data=json.dumps({'state': 'failed'}))
        data = json.loads(rv.data)

        self.assertEquals(data['status'], 403)
        self.assertEquals(data['error'], 'Forbidden')

    @patch('module_build_service.auth.get_user',
           return_value=('sammy', set(["packager", "mbs-admin"])))
    def test_cancel_build_admin(self, mocked_get_user):
        with patch("module_build_service.config.Config.admin_groups",
                   new_callable=PropertyMock, return_value=set(["mbs-admin"])):
            rv = self.client.patch('/module-build-service/1/module-builds/30',
                                   data=json.dumps({'state': 'failed'}))
            data = json.loads(rv.data)

            self.assertEquals(data['state'], 4)
            self.assertEquals(data['state_reason'], 'Canceled by sammy.')

    @patch('module_build_service.auth.get_user',
           return_value=('sammy', set(["packager"])))
    def test_cancel_build_no_admin(self, mocked_get_user):
        with patch("module_build_service.config.Config.admin_groups",
                   new_callable=PropertyMock, return_value=set(["mbs-admin"])):
            rv = self.client.patch('/module-build-service/1/module-builds/30',
                                   data=json.dumps({'state': 'failed'}))
            data = json.loads(rv.data)

            self.assertEquals(data['status'], 403)
            self.assertEquals(data['error'], 'Forbidden')

    @patch('module_build_service.auth.get_user', return_value=other_user)
    def test_cancel_build_wrong_param(self, mocked_get_user):
        rv = self.client.patch('/module-build-service/1/module-builds/30',
                               data=json.dumps({'some_param': 'value'}))
        data = json.loads(rv.data)

        self.assertEquals(data['status'], 400)
        self.assertEquals(data['error'], 'Bad Request')
        self.assertEquals(
            data['message'], 'Invalid JSON submitted')

    @patch('module_build_service.auth.get_user', return_value=other_user)
    def test_cancel_build_wrong_state(self, mocked_get_user):
        rv = self.client.patch('/module-build-service/1/module-builds/30',
                               data=json.dumps({'state': 'some_state'}))
        data = json.loads(rv.data)

        self.assertEquals(data['status'], 400)
        self.assertEquals(data['error'], 'Bad Request')
        self.assertEquals(
            data['message'], 'The provided state change is not supported')

    @patch('module_build_service.auth.get_user', return_value=user)
    def test_submit_build_unsupported_scm_scheme(self, mocked_get_user):
        scmurl = 'unsupported://example.com/modules/'
        'testmodule.git?#0000000000000000000000000000000000000000'
        rv = self.client.post('/module-build-service/1/module-builds/', data=json.dumps(
            {'branch': 'master', 'scmurl': scmurl}))
        data = json.loads(rv.data)
        self.assertIn(
            data['message'], (
                "The submitted scmurl {} is not allowed".format(scmurl),
                "The submitted scmurl {} is not valid".format(scmurl),
            )
        )
        self.assertEquals(data['status'], 403)
        self.assertEquals(data['error'], 'Forbidden')

    @patch('module_build_service.auth.get_user', return_value=user)
    @patch('module_build_service.scm.SCM')
    def test_submit_build_version_set_error(self, mocked_scm, mocked_get_user):
        FakeSCM(mocked_scm, 'testmodule', 'testmodule-version-set.yaml',
                '620ec77321b2ea7b0d67d82992dda3e1d67055b4')

        rv = self.client.post('/module-build-service/1/module-builds/', data=json.dumps(
            {'branch': 'master', 'scmurl': 'git://pkgs.stg.fedoraproject.org/modules/'
                'testmodule.git?#68931c90de214d9d13feefbd35246a81b6cb8d49'}))
        data = json.loads(rv.data)
        self.assertEquals(data['status'], 400)
        self.assertEquals(
            data['message'],
            'The version "123456789" is already defined in the modulemd but '
            'it shouldn\'t be since the version is generated based on the '
            'commit time')
        self.assertEquals(data['error'], 'Bad Request')

    @patch('module_build_service.auth.get_user', return_value=user)
    @patch('module_build_service.scm.SCM')
    def test_submit_build_wrong_stream(self, mocked_scm, mocked_get_user):
        FakeSCM(mocked_scm, 'testmodule', 'testmodule-wrong-stream.yaml',
                '620ec77321b2ea7b0d67d82992dda3e1d67055b4')

        rv = self.client.post('/module-build-service/1/module-builds/', data=json.dumps(
            {'branch': 'master', 'scmurl': 'git://pkgs.stg.fedoraproject.org/modules/'
                'testmodule.git?#68931c90de214d9d13feefbd35246a81b6cb8d49'}))
        data = json.loads(rv.data)
        self.assertEquals(data['status'], 400)
        self.assertEquals(
            data['message'],
            'The stream "wrong_stream" that is stored in the modulemd does not '
            'match the branch "master"')
        self.assertEquals(data['error'], 'Bad Request')

    @patch('module_build_service.auth.get_user', return_value=user)
    def test_submit_build_set_owner(self, mocked_get_user):
        data = {
            'branch': 'master',
            'scmurl': 'git://pkgs.stg.fedoraproject.org/modules/'
                      'testmodule.git?#68931c90de214d9d13feefbd35246a81b6cb8d49',
            'owner': 'foo',
        }
        rv = self.client.post('/module-build-service/1/module-builds/', data=json.dumps(data))
        result = json.loads(rv.data)
        self.assertEquals(result['status'], 400)
        self.assertIn("The request contains 'owner' parameter", result['message'])

    @patch('module_build_service.auth.get_user', return_value=anonymous_user)
    @patch('module_build_service.scm.SCM')
    @patch("module_build_service.config.Config.no_auth", new_callable=PropertyMock,
           return_value=True)
    def test_submit_build_no_auth_set_owner(self, mocked_conf, mocked_scm, mocked_get_user):
        FakeSCM(mocked_scm, 'testmodule', 'testmodule.yaml',
                '620ec77321b2ea7b0d67d82992dda3e1d67055b4')

        data = {
            'branch': 'master',
            'scmurl': 'git://pkgs.stg.fedoraproject.org/modules/'
                      'testmodule.git?#68931c90de214d9d13feefbd35246a81b6cb8d49',
            'owner': 'foo',
        }
        rv = self.client.post('/module-build-service/1/module-builds/', data=json.dumps(data))
        result = json.loads(rv.data)

        build = ModuleBuild.query.filter(ModuleBuild.id == result['id']).one()
        self.assertTrue(build.owner == result['owner'] == 'foo')

    @patch('module_build_service.auth.get_user', return_value=anonymous_user)
    @patch('module_build_service.scm.SCM')
    @patch("module_build_service.config.Config.no_auth", new_callable=PropertyMock)
    def test_patch_set_different_owner(self, mocked_no_auth, mocked_scm, mocked_get_user):
        FakeSCM(mocked_scm, 'testmodule', 'testmodule.yaml',
                '620ec77321b2ea7b0d67d82992dda3e1d67055b4')

        mocked_no_auth.return_value = True
        data = {
            'branch': 'master',
            'scmurl': 'git://pkgs.stg.fedoraproject.org/modules/'
                      'testmodule.git?#68931c90de214d9d13feefbd35246a81b6cb8d49',
            'owner': 'foo',
        }
        rv = self.client.post('/module-build-service/1/module-builds/', data=json.dumps(data))
        r1 = json.loads(rv.data)

        url = '/module-build-service/1/module-builds/' + str(r1['id'])
        r2 = self.client.patch(url, data=json.dumps({'state': 'failed'}))
        self.assertEquals(r2.status_code, 403)

        r3 = self.client.patch(url, data=json.dumps({'state': 'failed', 'owner': 'foo'}))
        self.assertEquals(r3.status_code, 200)

        mocked_no_auth.return_value = False
        r3 = self.client.patch(url, data=json.dumps({'state': 'failed', 'owner': 'foo'}))
        self.assertEquals(r3.status_code, 400)
        self.assertIn("The request contains 'owner' parameter", json.loads(r3.data)['message'])

    @patch('module_build_service.auth.get_user', return_value=user)
    @patch('module_build_service.scm.SCM')
    def test_submit_build_commit_hash_not_found(self, mocked_scm, mocked_get_user):
        FakeSCM(mocked_scm, 'testmodule', 'testmodule.yaml',
                '7035bd33614972ac66559ac1fdd019ff6027ad22', checkout_raise=True)

        rv = self.client.post('/module-build-service/1/module-builds/', data=json.dumps(
            {'branch': 'master', 'scmurl': 'git://pkgs.stg.fedoraproject.org/modules/'
                'testmodule.git?#7035bd33614972ac66559ac1fdd019ff6027ad22'}))
        data = json.loads(rv.data)
        self.assertIn("The requested commit hash was not found within the repository.",
                      data['message'])
        self.assertIn("Perhaps you forgot to push. The original message was: ",
                      data['message'])
        self.assertEquals(data['status'], 422)
        self.assertEquals(data['error'], 'Unprocessable Entity')

    @patch('module_build_service.auth.get_user', return_value=user)
    @patch('module_build_service.scm.SCM')
    @patch("module_build_service.config.Config.allow_custom_scmurls", new_callable=PropertyMock)
    def test_submit_custom_scmurl(self, allow_custom_scmurls, mocked_scm, mocked_get_user):
        FakeSCM(mocked_scm, 'testmodule', 'testmodule.yaml',
                '620ec77321b2ea7b0d67d82992dda3e1d67055b4')

        def submit(scmurl):
            return self.client.post('/module-build-service/1/module-builds/', data=json.dumps(
                                    {'branch': 'master', 'scmurl': scmurl}))

        allow_custom_scmurls.return_value = False
        res1 = submit('git://some.custom.url.org/modules/testmodule.git?#68931c9')
        data = json.loads(res1.data)
        self.assertEquals(data['status'], 403)
        self.assertTrue(data['message'].startswith('The submitted scmurl'))
        self.assertTrue(data['message'].endswith('is not allowed'))

        allow_custom_scmurls.return_value = True
        res2 = submit('git://some.custom.url.org/modules/testmodule.git?#68931c9')
        self.assertEquals(res2.status_code, 201)

    def test_about(self):
        with patch.object(mbs_config.Config, 'auth_method', new_callable=PropertyMock) as auth:
            auth.return_value = 'kerberos'
            rv = self.client.get('/module-build-service/1/about/')
        data = json.loads(rv.data)
        self.assertEqual(rv.status_code, 200)
        self.assertEquals(data, {'auth_method': 'kerberos', 'version': version})

    def test_rebuild_strategy_api(self):
        rv = self.client.get('/module-build-service/1/rebuild-strategies/')
        data = json.loads(rv.data)
        self.assertEqual(rv.status_code, 200)
        expected = {
            'items': [
                {
                    'allowed': False,
                    'default': False,
                    'description': 'All components will be rebuilt',
                    'name': 'all'
                },
                {
                    'allowed': True,
                    'default': True,
                    'description': ('All components that have changed and those in subsequent '
                                    'batches will be rebuilt'),
                    'name': 'changed-and-after'
                },
                {
                    'allowed': False,
                    'default': False,
                    'description': 'All changed components will be rebuilt',
                    'name': 'only-changed'
                }
            ]
        }
        self.assertEquals(data, expected)

    def test_rebuild_strategy_api_only_changed_default(self):
        with patch.object(mbs_config.Config, 'rebuild_strategy', new_callable=PropertyMock) as r_s:
            r_s.return_value = 'only-changed'
            rv = self.client.get('/module-build-service/1/rebuild-strategies/')
        data = json.loads(rv.data)
        self.assertEqual(rv.status_code, 200)
        expected = {
            'items': [
                {
                    'allowed': False,
                    'default': False,
                    'description': 'All components will be rebuilt',
                    'name': 'all'
                },
                {
                    'allowed': False,
                    'default': False,
                    'description': ('All components that have changed and those in subsequent '
                                    'batches will be rebuilt'),
                    'name': 'changed-and-after'
                },
                {
                    'allowed': True,
                    'default': True,
                    'description': 'All changed components will be rebuilt',
                    'name': 'only-changed'
                }
            ]
        }
        self.assertEquals(data, expected)

    def test_rebuild_strategy_api_override_allowed(self):
        with patch.object(mbs_config.Config, 'rebuild_strategy_allow_override',
                          new_callable=PropertyMock) as rsao:
            rsao.return_value = True
            rv = self.client.get('/module-build-service/1/rebuild-strategies/')
        data = json.loads(rv.data)
        self.assertEqual(rv.status_code, 200)
        expected = {
            'items': [
                {
                    'allowed': True,
                    'default': False,
                    'description': 'All components will be rebuilt',
                    'name': 'all'
                },
                {
                    'allowed': True,
                    'default': True,
                    'description': ('All components that have changed and those in subsequent '
                                    'batches will be rebuilt'),
                    'name': 'changed-and-after'
                },
                {
                    'allowed': True,
                    'default': False,
                    'description': 'All changed components will be rebuilt',
                    'name': 'only-changed'
                }
            ]
        }
        self.assertEquals(data, expected)



# -*- coding: utf-8 -*-


# Copyright (c) 2016  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Written by Lubos Kocman <lkocman@redhat.com>

"""PDC handler functions."""

import modulemd
from pdc_client import PDCClient
from module_build_service import db
from module_build_service import models
from module_build_service.errors import UnprocessableEntity

import inspect
import pprint
import logging
import six
import copy
import kobo.rpmlib
log = logging.getLogger()


def get_pdc_client_session(config):
    """
    :param config: instance of module_build_service.config.Config
    :return pdc_client.PDCClient instance
    """
    if 'ssl_verify' in inspect.getargspec(PDCClient.__init__).args:
        # New API
        return PDCClient(
            server=config.pdc_url,
            develop=config.pdc_develop,
            ssl_verify=not config.pdc_insecure,
        )
    else:
        # Old API
        return PDCClient(
            server=config.pdc_url,
            develop=config.pdc_develop,
            insecure=config.pdc_insecure,
        )


def get_variant_dict(data):
    """
    :param data: one of following
                    pdc variant_dict {'variant_id': value, 'variant_version': value, }
                    module dict {'name': value, 'version': value }
                    modulemd

    :return final list of module_info which pass repoclosure
    """
    def is_module_dict(data):
        if not isinstance(data, dict):
            return False

        for attr in ('name', 'version'):
            if attr not in data.keys():
                return False
        return True

    def is_variant_dict(data):
        if not isinstance(data, dict):
            return False

        if ('variant_id' not in data or
                ('variant_stream' not in data and
                 'variant_version' not in data)):
            return False
        return True

    def is_modulemd(data):
        return isinstance(data, modulemd.ModuleMetadata)

    def is_module_str(data):
        return isinstance(data, six.string_types)

    result = None

    if is_module_str(data):
        result = variant_dict_from_str(data)

    elif is_modulemd(data):
        result = {'variant_id': data.name}
        # Check if this is an old modulemd that doesn't use the new nomenclature
        if hasattr(data, 'release'):
            result['variant_release'] = data.release
            result['variant_version'] = data.version
        else:
            result['variant_release'] = data.version
            result['variant_version'] = data.stream

    elif is_variant_dict(data):
        result = data.copy()

        # This is a transitionary thing until we've ported PDC away from the old nomenclature
        if 'variant_version' not in result and 'variant_stream' in result:
            result['variant_version'] = result['variant_stream']
            del result['variant_stream']

        # ensure that variant_type is in result
        if 'variant_type' not in result.keys():
            result['variant_type'] = 'module'

    elif is_module_dict(data):
        result = {'variant_id': data['name'], 'variant_version': data['version']}

        if 'release' in data:
            result['variant_release'] = data['release']

        if 'active' in data:
            result['active'] = data['active']

    if not result:
        raise ValueError("Couldn't get variant_dict from %s" % data)

    return result


def variant_dict_from_str(module_str):
    """
    :param module_str: a string to match in PDC
    :return module_info dict

    Example minimal module_info:
        {
            'variant_id': module_name,
            'variant_version': module_version,
            'variant_type': 'module'
        }
    """
    log.debug("variant_dict_from_str(%r)" % module_str)
    # best match due several filters not being provided such as variant type ...

    module_info = {}

    release_start = module_str.rfind('-')
    version_start = module_str.rfind('-', 0, release_start)
    module_info['variant_release'] = module_str[release_start + 1:]
    module_info['variant_version'] = module_str[version_start + 1:release_start]
    module_info['variant_id'] = module_str[:version_start]
    module_info['variant_type'] = 'module'

    return module_info


def get_module(session, module_info, strict=False):
    """
    :param session : PDCClient instance
    :param module_info: pdc variant_dict, str, mmd or module dict
    :param strict: Normally this function returns None if no module can be
           found.  If strict=True, then an UnprocessableEntity is raised.

    :return final list of module_info which pass repoclosure
    """

    log.debug("get_module(%r, strict=%r)" % (module_info, strict))
    variant_dict = get_variant_dict(module_info)

    query = dict(
        variant_id=variant_dict['variant_id'],
        variant_version=variant_dict['variant_version'],
    )
    if variant_dict.get('variant_release'):
        query['variant_release'] = variant_dict['variant_release']
    if module_info.get('active'):
        query['active'] = module_info['active']

    # TODO: So far sorting on Fedora prod PDC instance is broken and it sorts
    # only by variant_uid by default. Once the sorting is fixed, we can start
    # using '-variant_release' ordering and just get the first variant from
    # there. But in the meantime, we have to get the first variant with
    # page_size set to 1 to find out how many variants (pages) are there in
    # results set and jump to last one in another query. The last one is always
    # the latest one (the one with the highest version).
    try:
        retval = session['unreleasedvariants/'](page_size=1, **query)
    except Exception as ex:
        log.debug("error during PDC lookup: %r" % ex)
        raise RuntimeError("Error during PDC lookup for module %s" % module_info["name"])

    # Error handling
    if not retval or len(retval["results"]) == 0:
        if strict:
            raise UnprocessableEntity("Failed to find module in PDC %r" % query)
        else:
            return None

    # Jump to last page to latest module release.
    if retval['count'] != 1:
        query['page'] = retval['count']
        retval = session['unreleasedvariants/'](page_size=1, **query)

    results = retval["results"]
    assert len(results) <= 1, pprint.pformat(retval)
    return results[0]


def get_module_tag(session, module_info, strict=False):
    """
    :param session : PDCClient instance
    :param module_info: list of module_info dicts
    :param strict: Normally this function returns None if no module can be
           found.  If strict=True, then an UnprocessableEntity is raised.
    :return: koji tag string
    """
    return get_module(session, module_info, strict=strict)['koji_tag']


def get_module_modulemd(session, module_info, strict=False):
    """
    :param session : PDCClient instance
    :param module_info: list of module_info dicts
    :param strict: Normally this function returns None if no module can be
           found.  If strict=True, then an UnprocessableEntity is raised.
    :return: ModuleMetadata instance
    """
    yaml = None
    module = get_module(session, module_info, strict=strict)
    if module:
        yaml = module['modulemd']

    if not yaml:
        if strict:
            raise UnprocessableEntity(
                "Failed to find modulemd entry in PDC for %r" % module_info)
        else:
            return None

    return _extract_modulemd(yaml, strict=strict)


def _extract_modulemd(yaml, strict=False):
    mmd = modulemd.ModuleMetadata()
    mmd.loads(yaml)
    return mmd


def _get_recursively_required_modules(session, info, modules=None,
                                      strict=False):
    """
    :param session: PDCClient instance
    :param info: pdc variant_dict, str, mmd or module dict
    :param modules: Used by recursion only, list of modules found by previous
                    iteration of this method.
    :param strict: Normally this function returns empty list if no module can
                   be found.  If strict=True, then an UnprocessableEntity is raised.

    Returns list of modules by recursively querying PDC based on a "requires"
    list of an input module represented by `info`. The returned list
    therefore contains all modules the input module "requires".

    If there are some modules loaded by utils.load_local_builds(...), these
    local modules will be used instead of querying PDC for the particular
    module found in local module builds database.

    The returned list contains only "modulemd" and "koji_tag" fields returned
    by PDC, because other fields are not known for local builds.
    """
    modules = modules or []

    variant_dict = get_variant_dict(info)
    local_modules = models.ModuleBuild.local_modules(
        db.session, variant_dict["variant_id"],
        variant_dict['variant_version'])
    if local_modules:
        local_module = local_modules[0]
        log.info("Using local module %r as a dependency.",
                 local_module)
        mmd = local_module.mmd()
        module_info = {}
        module_info["modulemd"] = local_module.modulemd
        module_info["koji_tag"] = local_module.koji_tag
    else:
        module_info = get_module(session, variant_dict, strict)
        module_info = {k: v for k, v in module_info.items()
                       if k in ["modulemd", "koji_tag"]}
        module_info = {
            'modulemd': module_info['modulemd'],
            'koji_tag': module_info['koji_tag']
        }

        yaml = module_info['modulemd']
        mmd = _extract_modulemd(yaml)

    # Check if we have examined this koji_tag already - no need to do
    # it again...
    if module_info in modules:
        return modules

    modules.append(module_info)

    # We want to use the same stream as the one used in the time this
    # module was built. But we still should fallback to plain mmd.requires
    # in case this module depends on some older module for which we did
    # not populate mmd.xmd['mbs']['requires'].
    if mmd.xmd.get('mbs') and mmd.xmd['mbs'].get('requires'):
        requires = mmd.xmd['mbs']['requires']
        requires = {name: data['stream'] for name, data in requires.items()}
    else:
        requires = mmd.requires

    for name, stream in requires.items():
        modified_dep = {
            'name': name,
            'version': stream,
            # Only return details about module builds that finished
            'active': True,
        }
        modules = _get_recursively_required_modules(
            session, modified_dep, modules, strict)

    return modules


def resolve_profiles(session, mmd, keys):
    """
    :param session : PDCClient instance
    :param mmd: ModuleMetadata instance of module
    :param keys: list of modulemd installation profiles to include in
                 the result.
    :return: Dictionary with keys set according to `keys` param and values
             set to union of all components defined in all installation
             profiles matching the key using the buildrequires.

    If there are some modules loaded by utils.load_local_builds(...), these
    local modules will be considered when returning the profiles.

    https://pagure.io/fm-orchestrator/issue/181
    """

    results = {}
    for key in keys:
        results[key] = set()
    for module_name, module_info in mmd.xmd['mbs']['buildrequires'].items():
        local_modules = models.ModuleBuild.local_modules(
            db.session, module_name, module_info['stream'])
        if local_modules:
            local_module = local_modules[0]
            log.info("Using local module %r to resolve profiles.",
                     local_module)
            dep_mmd = local_module.mmd()
            for key in keys:
                if key in dep_mmd.profiles:
                    results[key] |= dep_mmd.profiles[key].rpms
            continue

        # Find the dep in the built modules in PDC
        module_info = {
            'variant_id': module_name,
            'variant_stream': module_info['stream'],
            'variant_release': module_info['version']}
        modules = _get_recursively_required_modules(
            session, module_info, strict=True)

        for module in modules:
            yaml = module['modulemd']
            dep_mmd = _extract_modulemd(yaml)

            # Take note of what rpms are in this dep's profile.
            for key in keys:
                if key in dep_mmd.profiles:
                    results[key] |= dep_mmd.profiles[key].rpms

    # Return the union of all rpms in all profiles of the given keys.
    return results


def get_module_build_dependencies(session, module_info, strict=False):
    """
    :param session : PDCClient instance
    :param module_info : a dict containing filters for pdc or ModuleMetadata
    instance.
    :param strict: Normally this function returns None if no module can be
           found.  If strict=True, then an UnprocessableEntity is raised.
    :return dict with koji_tag as a key and ModuleMetadata object as value.

    Example minimal module_info:
        {
            'variant_id': module_name,
            'variant_version': module_version,
            'variant_type': 'module'
        }
    """
    log.debug("get_module_build_dependencies(%r, strict=%r)" % (module_info, strict))
    # XXX get definitive list of modules

    # This is the set we're going to build up and return.
    module_tags = {}

    if not isinstance(module_info, modulemd.ModuleMetadata):
        queried_module = get_module(session, module_info, strict=strict)
        yaml = queried_module['modulemd']
        queried_mmd = _extract_modulemd(yaml, strict=strict)
    else:
        queried_mmd = module_info

    if (not queried_mmd or not queried_mmd.xmd.get('mbs') or
            not queried_mmd.xmd['mbs'].get('buildrequires')):
        raise RuntimeError(
            'The module "{0!r}" did not contain its modulemd or did not have '
            'its xmd attribute filled out in PDC'.format(module_info))

    buildrequires = queried_mmd.xmd['mbs']['buildrequires']
    # Queue up the next tier of deps that we should look at..
    for name, details in buildrequires.items():
        modified_dep = {
            'name': name,
            'version': details['stream'],
            'release': details['version'],
            # Only return details about module builds that finished
            'active': True,
        }
        modules = _get_recursively_required_modules(
            session, modified_dep, strict=strict)
        for m in modules:
            if m["koji_tag"] in module_tags:
                continue
            module_tags[m["koji_tag"]] = _extract_modulemd(m["modulemd"])

    return module_tags


def resolve_requires(session, requires):
    """
    Takes `requires` dict with module_name as key and module_stream as value.
    Resolves the stream to particular latest version of a module and returns
    new dict in following format:

    {
        "module_name": {
            "ref": module_commit_hash,
            "stream": original_module_stream,
            "version": module_version,
            "filtered_rpms": ["nvr", ...]
        },
        ...
    }

    If there are some modules loaded by utils.load_local_builds(...), these
    local modules will be considered when resolving the requires.

    Raises RuntimeError on PDC lookup error.
    """
    new_requires = copy.deepcopy(requires)
    for module_name, module_stream in requires.items():
        # Try to find out module dependency in the local module builds
        # added by utils.load_local_builds(...).
        local_modules = models.ModuleBuild.local_modules(
            db.session, module_name, module_stream)
        if local_modules:
            local_build = local_modules[0]
            new_requires[module_name] = {
                # The commit ID isn't currently saved in modules.yaml
                'ref': None,
                'stream': local_build.stream,
                'version': local_build.version,
                # No need to set filtered_rpms for local builds, because MBS
                # filters the RPMs automatically when the module build is
                # done.
                'filtered_rpms': []
            }
            continue

        # Assumes that module_stream is the stream and not the commit hash
        module_info = {
            'name': module_name,
            'version': module_stream,
            'active': True}
        module = get_module(session, module_info)

        commit_hash = None
        version = None
        filtered_rpms = []
        module = get_module(session, module_info, strict=True)
        if module.get('modulemd'):
            mmd = modulemd.ModuleMetadata()
            mmd.loads(module['modulemd'])
            if mmd.xmd.get('mbs') and mmd.xmd['mbs'].get('commit'):
                commit_hash = mmd.xmd['mbs']['commit']

            # Find out the particular NVR of filtered packages
            if "rpms" in module and mmd.filter and mmd.filter.rpms:
                for rpm in module["rpms"]:
                    nvr = kobo.rpmlib.parse_nvra(rpm)
                    # If the package is not filtered, continue
                    if not nvr["name"] in mmd.filter.rpms:
                        continue

                    # If the nvr is already in filtered_rpms, continue
                    nvr = kobo.rpmlib.make_nvr(nvr, force_epoch=True)
                    if nvr in filtered_rpms:
                        continue
                    filtered_rpms.append(nvr)

        if module.get('variant_release'):
            version = module['variant_release']

        if version and commit_hash:
            new_requires[module_name] = {
                'ref': commit_hash,
                'stream': module_stream,
                'version': str(version),
                'filtered_rpms': filtered_rpms,
            }
        else:
            raise RuntimeError(
                'The module "{0}" didn\'t contain either a commit hash or a'
                ' version in PDC'.format(module_name))

    return new_requires



"""state reason history

Revision ID: 335455a30585
Revises: 474697622859
Create Date: 2017-02-10 11:38:55.249234

"""

# revision identifiers, used by Alembic.
revision = '335455a30585'
down_revision = '474697622859'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.create_table(
        'module_builds_trace',
        sa.Column('id', sa.Integer(), primary_key=True),
        sa.Column('module_id', sa.Integer(), sa.ForeignKey('module_builds.id'), nullable=False),
        sa.Column('state_time', sa.DateTime(), nullable=False),
        sa.Column('state', sa.Integer(), nullable=True),
        sa.Column('state_reason', sa.String(), nullable=True)
    )

    op.create_table(
        'component_builds_trace',
        sa.Column('id', sa.Integer(), primary_key=True),
        sa.Column('component_id', sa.Integer(), sa.ForeignKey('component_builds.id'), nullable=False),
        sa.Column('state_time', sa.DateTime(), nullable=False),
        sa.Column('state', sa.Integer(), nullable=True),
        sa.Column('state_reason', sa.String(), nullable=True),
        sa.Column('task_id', sa.Integer(), nullable=True),
    )


def downgrade():
    op.drop_table('module_builds_trace')
    op.drop_table('component_builds_trace')



# -*- coding: utf-8 -*-
# Copyright (c) 2016  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Written by Petr Šabata <contyk@redhat.com>
#            Matt Prahl <mprahl@redhat.com>

""" The module build orchestrator for Modularity, API.
This is the implementation of the orchestrator's public RESTful API.
"""

import json
import module_build_service.auth
from flask import request, jsonify
from flask.views import MethodView

from module_build_service import app, conf, log, models, db, version
from module_build_service.utils import (
    pagination_metadata, filter_module_builds, filter_component_builds,
    submit_module_build_from_scm, submit_module_build_from_yaml,
    get_scm_url_re)
from module_build_service.errors import (
    ValidationError, Forbidden, NotFound, ProgrammingError)

api_v1 = {
    'module_builds': {
        'url': '/module-build-service/1/module-builds/',
        'options': {
            'methods': ['POST'],
        }
    },
    'module_builds_list': {
        'url': '/module-build-service/1/module-builds/',
        'options': {
            'defaults': {'id': None},
            'methods': ['GET'],
        }
    },
    'module_build': {
        'url': '/module-build-service/1/module-builds/<int:id>',
        'options': {
            'methods': ['GET', 'PATCH'],
        }
    },
    'component_builds_list': {
        'url': '/module-build-service/1/component-builds/',
        'options': {
            'defaults': {'id': None},
            'methods': ['GET'],
        }
    },
    'component_build': {
        'url': '/module-build-service/1/component-builds/<int:id>',
        'options': {
            'methods': ['GET'],
        }
    },
    'about': {
        'url': '/module-build-service/1/about/',
        'options': {
            'methods': ['GET']
        }
    },
    'rebuild_strategies_list': {
        'url': '/module-build-service/1/rebuild-strategies/',
        'options': {
            'methods': ['GET']
        }
    }
}


class AbstractQueryableBuildAPI(MethodView):
    """ An abstract class, housing some common functionality. """

    def get(self, id):
        verbose_flag = request.args.get('verbose', 'false').lower()

        if id is None:
            # Lists all tracked builds
            p_query = self.query_filter(request)

            json_data = {
                'meta': pagination_metadata(p_query, request.args)
            }

            if verbose_flag == 'true' or verbose_flag == '1':
                json_data['items'] = [item.extended_json() for item in p_query.items]
            else:
                json_data['items'] = [item.json() for item in p_query.items]

            return jsonify(json_data), 200
        else:
            # Lists details for the specified build
            instance = self.model.query.filter_by(id=id).first()
            if instance:
                if verbose_flag == 'true' or verbose_flag == '1':
                    return jsonify(instance.extended_json()), 200
                else:
                    return jsonify(instance.json()), 200
            else:
                raise NotFound('No such %s found.' % self.kind)


class ComponentBuildAPI(AbstractQueryableBuildAPI):
    kind = 'component'
    query_filter = staticmethod(filter_component_builds)
    model = models.ComponentBuild


class ModuleBuildAPI(AbstractQueryableBuildAPI):
    kind = 'module'
    query_filter = staticmethod(filter_module_builds)
    model = models.ModuleBuild

    # Additional POST and DELETE handlers for modules follow.

    def post(self):
        if "multipart/form-data" in request.headers.get("Content-Type", ""):
            handler = YAMLFileHandler(request)
        else:
            handler = SCMHandler(request)

        if conf.no_auth is True and handler.username == "anonymous" and "owner" in handler.data:
            handler.username = handler.data["owner"]

        if conf.allowed_groups and not (conf.allowed_groups & handler.groups):
            raise Forbidden("%s is not in any of  %r, only %r" % (
                handler.username, conf.allowed_groups, handler.groups))

        handler.validate()
        module = handler.post()
        return jsonify(module.extended_json()), 201

    def patch(self, id):
        username, groups = module_build_service.auth.get_user(request)

        try:
            r = json.loads(request.get_data().decode("utf-8"))
        except Exception:
            log.error('Invalid JSON submitted')
            raise ValidationError('Invalid JSON submitted')

        if "owner" in r:
            if conf.no_auth is not True:
                raise ValidationError(("The request contains 'owner' parameter,"
                                       " however NO_AUTH is not allowed"))
            elif username == "anonymous":
                username = r["owner"]

        if conf.allowed_groups and not (conf.allowed_groups & groups):
            raise Forbidden("%s is not in any of  %r, only %r" % (
                username, conf.allowed_groups, groups))

        module = models.ModuleBuild.query.filter_by(id=id).first()
        if not module:
            raise NotFound('No such module found.')

        if module.owner != username and not (conf.admin_groups & groups):
            raise Forbidden('You are not owner of this build and '
                            'therefore cannot modify it.')

        if not r.get('state'):
            log.error('Invalid JSON submitted')
            raise ValidationError('Invalid JSON submitted')

        if r['state'] == 'failed' \
                or r['state'] == str(models.BUILD_STATES['failed']):
            module.transition(conf, models.BUILD_STATES["failed"],
                              "Canceled by %s." % username)
        else:
            log.error('The provided state change of "{}" is not supported'
                      .format(r['state']))
            raise ValidationError('The provided state change is not supported')
        db.session.add(module)
        db.session.commit()

        return jsonify(module.extended_json()), 200


class AboutAPI(MethodView):
    def get(self):
        json = {'version': version}
        config_items = ['auth_method']
        for item in config_items:
            config_item = getattr(conf, item)
            # All config items have a default, so if doesn't exist it is a programming error
            if not config_item:
                raise ProgrammingError(
                    'An invalid config item of "{0}" was specified'.format(item))
            json[item] = config_item
        return jsonify(json), 200


class RebuildStrategies(MethodView):
    def get(self):
        items = []
        # Sort the items list by name
        for strategy in sorted(models.ModuleBuild.rebuild_strategies.keys()):
            default = False
            if strategy == conf.rebuild_strategy:
                default = True
                allowed = True
            elif conf.rebuild_strategy_allow_override and \
                    strategy in conf.rebuild_strategies_allowed:
                allowed = True
            else:
                allowed = False
            items.append({
                'name': strategy,
                'description': models.ModuleBuild.rebuild_strategies[strategy],
                'allowed': allowed,
                'default': default
            })

        return jsonify({'items': items}), 200


class BaseHandler(object):
    def __init__(self, request):
        self.username, self.groups = module_build_service.auth.get_user(request)
        self.data = None

    @property
    def optional_params(self):
        return {k: v for k, v in self.data.items() if k not in ["owner", "scmurl", "branch"]}

    def validate_optional_params(self):
        forbidden_params = [k for k in self.data
                            if k not in models.ModuleBuild.__table__.columns and
                            k not in ["branch", "rebuild_strategy"]]
        if forbidden_params:
            raise ValidationError('The request contains unspecified parameters: {}'
                                  .format(", ".join(forbidden_params)))

        forbidden_params = [k for k in self.data if k.startswith("copr_")]
        if conf.system != "copr" and forbidden_params:
            raise ValidationError(('The request contains parameters specific to Copr builder:'
                                   ' {} even though {} is used')
                                  .format(", ".join(forbidden_params), conf.system))

        if not conf.no_auth and "owner" in self.data:
            raise ValidationError(("The request contains 'owner' parameter,"
                                   " however NO_AUTH is not allowed"))

        if not conf.rebuild_strategy_allow_override and 'rebuild_strategy' in self.data:
            raise ValidationError('The request contains the "rebuild_strategy" parameter but '
                                  'overriding the default isn\'t allowed')

        if 'rebuild_strategy' in self.data:
            if self.data['rebuild_strategy'] not in conf.rebuild_strategies_allowed:
                raise ValidationError(
                    'The rebuild method of "{0}" is not allowed. Choose from: {1}.'
                    .format(self.data['rebuild_strategy'],
                            ', '.join(conf.rebuild_strategies_allowed)))


class SCMHandler(BaseHandler):
    def __init__(self, request):
        super(SCMHandler, self).__init__(request)
        try:
            self.data = json.loads(request.get_data().decode("utf-8"))
        except Exception:
            log.error('Invalid JSON submitted')
            raise ValidationError('Invalid JSON submitted')

    def validate(self):
        if "scmurl" not in self.data:
            log.error('Missing scmurl')
            raise ValidationError('Missing scmurl')

        url = self.data["scmurl"]
        allowed_prefix = any(url.startswith(prefix) for prefix in conf.scmurls)
        if not conf.allow_custom_scmurls and not allowed_prefix:
            log.error("The submitted scmurl %r is not allowed" % url)
            raise Forbidden("The submitted scmurl %s is not allowed" % url)

        if not get_scm_url_re().match(url):
            log.error("The submitted scmurl %r is not valid" % url)
            raise Forbidden("The submitted scmurl %s is not valid" % url)

        if "branch" not in self.data:
            log.error('Missing branch')
            raise ValidationError('Missing branch')

        self.validate_optional_params()

    def post(self):
        url = self.data["scmurl"]
        branch = self.data["branch"]

        # python-modulemd expects this to be bytes, not unicode.
        if isinstance(branch, unicode):
            branch = branch.encode('utf-8')

        return submit_module_build_from_scm(self.username, url, branch,
                                            allow_local_url=False,
                                            optional_params=self.optional_params)


class YAMLFileHandler(BaseHandler):
    def __init__(self, request):
        if not conf.yaml_submit_allowed:
            raise Forbidden("YAML submission is not enabled")
        super(YAMLFileHandler, self).__init__(request)
        self.data = request.form.to_dict()

    def validate(self):
        if "yaml" not in request.files:
            log.error('Invalid file submitted')
            raise ValidationError('Invalid file submitted')
        self.validate_optional_params()

    def post(self):
        handle = request.files["yaml"]
        return submit_module_build_from_yaml(self.username, handle,
                                             optional_params=self.optional_params)


def register_api_v1():
    """ Registers version 1 of MBS API. """
    module_view = ModuleBuildAPI.as_view('module_builds')
    component_view = ComponentBuildAPI.as_view('component_builds')
    about_view = AboutAPI.as_view('about')
    rebuild_strategies_view = RebuildStrategies.as_view('rebuild_strategies')
    for key, val in api_v1.items():
        if key.startswith('component_build'):
            app.add_url_rule(val['url'],
                             endpoint=key,
                             view_func=component_view,
                             **val['options'])
        elif key.startswith('module_build'):
            app.add_url_rule(val['url'],
                             endpoint=key,
                             view_func=module_view,
                             **val['options'])
        elif key.startswith('about'):
            app.add_url_rule(val['url'],
                             endpoint=key,
                             view_func=about_view,
                             **val['options'])
        elif key == 'rebuild_strategies_list':
            app.add_url_rule(val['url'],
                             endpoint=key,
                             view_func=rebuild_strategies_view,
                             **val['options'])
        else:
            raise NotImplementedError("Unhandled api key.")


register_api_v1()



from module_build_service import conf
from base import GenericBuilder
from KojiModuleBuilder import KojiModuleBuilder
from MockModuleBuilder import MockModuleBuilder

__all__ = [
    GenericBuilder
]


GenericBuilder.register_backend_class(KojiModuleBuilder)

GenericBuilder.register_backend_class(MockModuleBuilder)

if conf.system == "copr":
    from CoprModuleBuilder import CoprModuleBuilder
    GenericBuilder.register_backend_class(CoprModuleBuilder)
